import psycopg2
import os
import io
import sys
import json
import datetime
import decimal
import time
import binascii
class pg_encoder(json.JSONEncoder):
	def default(self, obj):
		if 	isinstance(obj, datetime.time) or \
			isinstance(obj, datetime.datetime) or  \
			isinstance(obj, datetime.date) or \
			isinstance(obj, decimal.Decimal) or \
			isinstance(obj, datetime.timedelta) or \
			isinstance(obj, set):
				
			return str(obj)
		return json.JSONEncoder.default(self, obj)


class pg_connection(object):
	def __init__(self, global_config):
		self.global_conf=global_config
		self.pg_conn=self.global_conf.pg_conn
		self.pg_database=self.global_conf.pg_database
		self.dest_schema=self.global_conf.my_database
		self.pg_connection=None
		self.pgsql_cur=None
		self.pgsql_cur_replay=None
		self.pg_charset=self.global_conf.pg_charset
		
	
	def connect_db(self):
		"""
			Connects to PostgreSQL using the parameters stored in pg_pars built adding the key dbname to the self.pg_conn dictionary.
			This method's connection and cursors are widely used in the procedure except for the replay process which uses a 
			dedicated connection and cursor.
			The method after the connection creates a database cursor and set the session to autocommit.
		"""
		pg_pars=dict(list(self.pg_conn.items())+ list({'dbname':self.pg_database}.items()))
		strconn="dbname=%(dbname)s user=%(user)s host=%(host)s password=%(password)s port=%(port)s"  % pg_pars
		self.pgsql_conn = psycopg2.connect(strconn)
		self.pgsql_conn .set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
		self.pgsql_conn .set_client_encoding(self.pg_charset)
		self.pgsql_cur=self.pgsql_conn .cursor()
	
	def connect_replay_db(self):
		"""
			Connects to PostgreSQL using the parameters stored in pg_pars built adding the key dbname to the self.pg_conn dictionary.
			The method after the connection creates a database cursor and set the session to autocommit.
			This method creates an additional connection and cursor used by the replay process. 

		"""
		pg_pars=dict(list(self.pg_conn.items())+ list({'dbname':self.pg_database}.items()))
		strconn="dbname=%(dbname)s user=%(user)s host=%(host)s password=%(password)s port=%(port)s"  % pg_pars
		self.pgsql_conn_replay = psycopg2.connect(strconn)
		self.pgsql_conn_replay.set_isolation_level(psycopg2.extensions.ISOLATION_LEVEL_AUTOCOMMIT)
		self.pgsql_conn_replay.set_client_encoding(self.pg_charset)
		self.pgsql_cur_replay=self.pgsql_conn_replay.cursor()
		
		
	def disconnect_db(self):
		"""
			The method disconnects from the main database connection.
		"""
		self.pgsql_conn.close()
	
	def disconnect_replay_db(self):
		"""
			The method disconnects from the replay database connection.
		"""
		self.pgsql_conn_replay.close()
		
	
		

class pg_engine(object):
	"""
		The class pg_engine manages the replica initialisation and execution on the PostgreSQL side.
		
		The class connects to the database when instantiated and setup several class attributes used by the replica.
		In particular the class dictionary type_dictionary is used to map the MySQL types to the equivalent PostgreSQL types.
		Unlike pgloader, which allows the type mapping configuration, the dictionary is hardcoded as the mapping is an effort to keep the replica running as smooth as possible.
		The class manages the replica catalogue upgrade using the current catalogue version self.cat_version and the list of migrations self.cat_sql.
		
		If the catalogue version, stored in sch_chameleon.v_version is different from the value stored in self.cat_version then the method upgrade_service_schema() is executed.
		
	"""
	def __init__(self, global_config, table_metadata, table_file, logger, sql_dir='sql/'):
		self.sleep_on_reindex = global_config.sleep_on_reindex
		self.reindex_app_names = global_config.reindex_app_names
		self.batch_retention = global_config.batch_retention
		self.type_override = global_config.type_override
		self.logger = logger
		self.sql_dir = sql_dir
		self.idx_sequence = 0
		self.pg_conn = pg_connection(global_config)
		self.pg_conn.connect_db()
		self.table_metadata = table_metadata
		self.table_file = table_file
		self.type_dictionary = {
			'integer':'integer',
			'mediumint':'bigint',
			'tinyint':'integer',
			'smallint':'integer',
			'int':'integer',
			'bigint':'bigint',
			'varchar':'character varying',
			'character varying':'character varying',
			'text':'text',
			'char':'character',
			'datetime':'timestamp without time zone',
			'date':'date',
			'time':'time without time zone',
			'timestamp':'timestamp without time zone',
			'tinytext':'text',
			'mediumtext':'text',
			'longtext':'text',
			'tinyblob':'bytea',
			'mediumblob':'bytea',
			'longblob':'bytea',
			'blob':'bytea', 
			'binary':'bytea', 
			'varbinary':'bytea', 
			'decimal':'numeric', 
			'double':'double precision', 
			'double precision':'double precision', 
			'float':'double precision', 
			'bit':'integer', 
			'year':'integer', 
			'enum':'enum', 
			'set':'text', 
			'json':'text', 
			'bool':'boolean', 
			'boolean':'boolean', 
			'geometry':'bytea',
		}
		self.table_ddl = {}
		self.idx_ddl = {}
		self.type_ddl = {}
		self.pg_charset = self.pg_conn.pg_charset
		self.cat_version = '1.7'
		self.cat_sql = [
			{'version':'base','script': 'create_schema.sql'}, 
			{'version':'0.1','script': 'upgrade/cat_0.1.sql'}, 
			{'version':'0.2','script': 'upgrade/cat_0.2.sql'}, 
			{'version':'0.3','script': 'upgrade/cat_0.3.sql'}, 
			{'version':'0.4','script': 'upgrade/cat_0.4.sql'}, 
			{'version':'0.5','script': 'upgrade/cat_0.5.sql'}, 
			{'version':'0.6','script': 'upgrade/cat_0.6.sql'}, 
			{'version':'0.7','script': 'upgrade/cat_0.7.sql'}, 
			{'version':'0.8','script': 'upgrade/cat_0.8.sql'}, 
			{'version':'0.9','script': 'upgrade/cat_0.9.sql'}, 
			{'version':'1.0','script': 'upgrade/cat_1.0.sql'}, 
			{'version':'1.1','script': 'upgrade/cat_1.1.sql'}, 
			{'version':'1.2','script': 'upgrade/cat_1.2.sql'}, 
			{'version':'1.3','script': 'upgrade/cat_1.3.sql'}, 
			{'version':'1.4','script': 'upgrade/cat_1.4.sql'},
			{'version':'1.5','script': 'upgrade/cat_1.5.sql'},
			{'version':'1.6','script': 'upgrade/cat_1.6.sql'},
			{'version':'1.7','script': 'upgrade/cat_1.7.sql'},
		]
		cat_version=self.get_schema_version()
		num_schema=(self.check_service_schema())[0]
		if cat_version!=self.cat_version and int(num_schema)>0:
			self.upgrade_service_schema()
		self.table_limit = ['*']
		self.master_status = None
	
	def set_application_name(self, action="", conn_type='main'):
		"""
			The method sets the application name in the replica using the variable self.pg_conn.global_conf.source_name,
			Making simpler to find the replication processes. If the source name is not set then a generic PGCHAMELEON name is used.
		"""
		if self.pg_conn.global_conf.source_name:
			app_name = "[PGCH] - source: %s, action: %s" % (self.pg_conn.global_conf.source_name, action)
		else:
			app_name = "[PGCH]"
		sql_app_name="""SET application_name=%s; """
		if conn_type == 'main':
			self.pg_conn.pgsql_cur.execute(sql_app_name, (app_name , ))
		elif conn_type == 'replay':
			self.pg_conn.pgsql_cur_replay.execute(sql_app_name, (app_name , ))
	
		
	def add_source(self, source_name, dest_schema):
		"""
			The method add a new source in the replica catalogue. 
			If the source name is already present an error message is emitted without further actions.
			:param source_name: The source name stored in the configuration parameter source_name.
			:param source_schema: The source schema on mysql. The field is not used except when migrating the catalogue to the newer version 2.0.x.
		"""
		sql_source = """
			SELECT 
				count(i_id_source)
			FROM 
				sch_chameleon.t_sources 
			WHERE 
				t_source=%s
				;
			"""
		self.pg_conn.pgsql_cur.execute(sql_source, (source_name, ))
		source_data = self.pg_conn.pgsql_cur.fetchone()
		cnt_source = source_data[0]
		if cnt_source == 0:
			sql_add = """
				INSERT INTO sch_chameleon.t_sources 
					( 
						t_source,
						t_dest_schema,
						t_source_schema
					) 
				VALUES 
					(
						%s,
						%s,
						%s
					)
				RETURNING 
					i_id_source
				; 
			"""
			self.pg_conn.pgsql_cur.execute(sql_add, (source_name, dest_schema, self.source_schema ))
			source_add = self.pg_conn.pgsql_cur.fetchone()
			sql_update = """
				UPDATE sch_chameleon.t_sources
					SET v_log_table=ARRAY[
						't_log_replica_1_src_%s',
						't_log_replica_2_src_%s'
					]
				WHERE i_id_source=%s
				;
			"""
			self.pg_conn.pgsql_cur.execute(sql_update,  (source_add[0],source_add[0], source_add[0] ))
			
			sql_parts = """SELECT sch_chameleon.fn_refresh_parts() ;"""
			self.pg_conn.pgsql_cur.execute(sql_parts)
			
		else:
			print("Source %s already registered." % source_name)
		sys.exit()
	
	def get_source_status(self, source_name):
		"""
			Gets the source status usin the source name.
			Possible values are:

			ready : the source is registered but the init_replica is not yet done.
			
			initialising: init_replica is initialising
			
			initialised: init_replica finished and the replica process is ready to start
			
			stopped: the replica process is stopped
			
			running: the replica process is running
				
			:param source_name: The source name stored in the configuration parameter source_name.
			:type source_name: string
			:return: source_status extracted from PostgreSQL
			:rtype: string
		"""
		sql_source = """
					SELECT 
						enm_status
					FROM 
						sch_chameleon.t_sources 
					WHERE 
						t_source=%s
				;
			"""
		self.pg_conn.pgsql_cur.execute(sql_source, (source_name, ))
		source_data = self.pg_conn.pgsql_cur.fetchone()
		if source_data:
			source_status = source_data[0]
		else:
			source_status = 'Not registered'
		return source_status
		
	def drop_source(self, source_name):
		"""
			Drops the source from the replication catalogue discarding any replica reference.
			:param source_name: The source name stored in the configuration parameter source_name.
		"""
		sql_delete = """ DELETE FROM sch_chameleon.t_sources 
					WHERE  t_source=%s
					RETURNING v_log_table
					; """
		self.pg_conn.pgsql_cur.execute(sql_delete, (source_name, ))
		source_drop = self.pg_conn.pgsql_cur.fetchone()
		for log_table in source_drop[0]:
			sql_drop = """DROP TABLE sch_chameleon."%s"; """ % (log_table)
			self.pg_conn.pgsql_cur.execute(sql_drop)
	
		
	
	def set_source_id(self, source_status):
		"""
			Sets the source status for the source_name and sets the two class attributes i_id_source and dest_schema.
			
			:param source_status: The source status to be set.
			
		"""
		sql_source = """
			UPDATE sch_chameleon.t_sources
			SET
				enm_status=%s,
				t_source_schema=%s
			WHERE
				t_source=%s
			RETURNING i_id_source,t_dest_schema
				;
			"""
		source_name = self.pg_conn.global_conf.source_name
		self.pg_conn.pgsql_cur.execute(sql_source, (source_status, self.source_schema,  source_name))
		source_data = self.pg_conn.pgsql_cur.fetchone()
		try:
			self.i_id_source = source_data[0]
			self.dest_schema = source_data[1]
			self.source_name = source_name
		except:
			print("Source %s is not registered." % source_name)
			sys.exit()
	
			
	def clean_batch_data(self):
		"""
			Removes the replica batch data for the given source id.
			The method is used to cleanup incomplete batch data in case of crash or replica's unclean restart
		"""
		sql_delete = """
			DELETE FROM sch_chameleon.t_replica_batch 
			WHERE i_id_source=%s;
		"""
		self.pg_conn.pgsql_cur.execute(sql_delete, (self.i_id_source, ))
		
		
	def create_schema(self):
		"""
			The method drops and creates the destination schema.
			It also set the search_path for the cursor to the destination schema.
		"""
		sql_drop="""DROP SCHEMA IF EXISTS "%s" CASCADE;""" % self.dest_schema
		sql_create="""CREATE SCHEMA IF NOT EXISTS "%s";""" % self.dest_schema
		self.pg_conn.pgsql_cur.execute(sql_drop)
		self.pg_conn.pgsql_cur.execute(sql_create)
		self.set_search_path()
	
	def store_table(self, table_name):
		"""
			The method saves the table name along with the primary key definition in the table t_replica_tables.
			This is required in order to let the replay procedure which primary key to use replaying the update and delete.
			If the table is without primary key is not stored. 
			A table without primary key is copied and the indices are create like any other table. 
			However the replica doesn't work for the tables without primary key.
			
			If the class variable master status is set then the master's coordinates are saved along with the table.
			This happens in general when a table is added to the replica or the data is refreshed with sync_tables.
			
			:param table_name: the table name to store in the table  t_replica_tables
		"""
		if self.master_status:
			master_data = self.master_status[0]
			binlog_file = master_data["File"]
			binlog_pos = master_data["Position"]
		else:
			binlog_file = None
			binlog_pos = None
		table_data=self.table_metadata[table_name]
		table_no_pk = True
		for index in table_data["indices"]:
			if index["index_name"]=="PRIMARY":
				table_no_pk = False
				sql_insert=""" 
					INSERT INTO sch_chameleon.t_replica_tables 
						(
							i_id_source,
							v_table_name,
							v_schema_name,
							v_table_pkey,
							t_binlog_name,
							i_binlog_position
						)
					VALUES 
						(
							%s,
							%s,
							%s,
							ARRAY[%s],
							%s,
							%s
						)
					ON CONFLICT (i_id_source,v_table_name,v_schema_name)
						DO UPDATE 
							SET 
								v_table_pkey=EXCLUDED.v_table_pkey,
								t_binlog_name = EXCLUDED.t_binlog_name,
								i_binlog_position = EXCLUDED.i_binlog_position
										;
								"""
				self.pg_conn.pgsql_cur.execute(sql_insert, (
					self.i_id_source, 
					table_name, 
					self.dest_schema, 
					index["index_columns"].strip(), 
					binlog_file, 
					binlog_pos
					)
				)
		if table_no_pk:
			self.logger.warning("Missing primary key. The table %s will not be replicated." % (table_name,))
			sql_delete = """
				DELETE FROM sch_chameleon.t_replica_tables
				WHERE
						i_id_source=%s
					AND	v_table_name=%s
					AND	v_schema_name=%s
				;
			"""
			self.pg_conn.pgsql_cur.execute(sql_delete, (
				self.i_id_source, 
				table_name, 
				self.dest_schema)
				)
		
	def unregister_table(self, table_name):
		"""
			This method is used when a table have the primary key dropped on MySQL. 
			The table name is removed from the replicatoin catalogue and the table is renamed.
			This way any dependency (e.g. views, functions) to the table is preserved but the replica is stopped.

			:param table_name: the table name to remove from t_replica_tables
		"""
		self.logger.info("unregistering table %s from the replica catalog" % (table_name,))
		sql_delete=""" DELETE FROM sch_chameleon.t_replica_tables 
									WHERE
											v_table_name=%s
										AND	v_schema_name=%s
								RETURNING i_id_table
								;
						"""
		self.pg_conn.pgsql_cur.execute(sql_delete, (table_name, self.dest_schema))	
		removed_id=self.pg_conn.pgsql_cur.fetchone()
		table_id=removed_id[0]
		self.logger.info("renaming table %s to %s_%s" % (table_name, table_name, table_id))
		sql_rename="""ALTER TABLE IF EXISTS "%s"."%s" rename to "%s_%s"; """ % (self.dest_schema, table_name, table_name, table_id)
		self.logger.debug(sql_rename)
		self.pg_conn.pgsql_cur.execute(sql_rename)	
	
	
	def set_search_path(self):
		"""
			The method sets the search path for the connection.
		"""
		sql_path=""" SET search_path="%s";""" % (self.dest_schema, )
		self.pg_conn.pgsql_cur.execute(sql_path)
		
	
	def drop_tables(self):
		"""
			The method drops the tables present in the table_ddl
		"""
		self.set_search_path()
		for table in self.table_ddl:
			self.logger.debug("dropping table %s " % (table, ))
			sql_drop = """DROP TABLE IF EXISTS "%s"  CASCADE;""" % (table, )
			self.pg_conn.pgsql_cur.execute(sql_drop)
			
	
	def create_tables(self):
		"""
			The method loops trough the list table_ddl and executes the creation scripts.
			No index is created in this method
		"""
		self.set_search_path()
		for table in self.table_ddl:
			self.logger.debug("creating table %s " % (table, ))
			try:
				ddl_enum=self.type_ddl[table]
				for sql_type in ddl_enum:
					self.pg_conn.pgsql_cur.execute(sql_type)
			except psycopg2.Error as e:
				self.logger.error("SQLCODE: %s SQLERROR: %s" % (e.pgcode, e.pgerror))
				self.logger.error(sql_type)
				
			sql_create=self.table_ddl[table]
			try:
				self.pg_conn.pgsql_cur.execute(sql_create)
			except psycopg2.Error as e:
				self.logger.error("SQLCODE: %s SQLERROR: %s" % (e.pgcode, e.pgerror))
				self.logger.error(sql_create)
				
			self.store_table(table)

	def create_indices(self):
		"""
			The method creates the indices using the DDL stored in the class list self.idx_ddl.
		"""
		self.logger.info("creating the indices")
		for table in self.idx_ddl:
			idx_ddl= self.idx_ddl[table]
			self.logger.debug("processing table %s" % (table, ))
			for sql_idx in idx_ddl:
				self.pg_conn.pgsql_cur.execute(sql_idx)
				
	
	def copy_data(self, table,  csv_file,  my_tables={}):
		"""
			The method copy the data into postgresql using psycopg2's copy_expert.
			The csv_file is a file like object which can be either a  csv file or a string io object, accordingly with the 
			configuration parameter copy_mode.
			
			:param table: the table name, used to get the table's metadata out of my_tables
			:param csv_file: file like object with the table's data stored in CSV format
			:param my_tables: table's metadata dictionary 
		"""
		column_copy=[]
		for column in my_tables[table]["columns"]:
			column_copy.append('"'+column["column_name"]+'"')
		sql_copy="COPY "+'"'+self.dest_schema+'"'+"."+'"'+table+'"'+" ("+','.join(column_copy)+") FROM STDIN WITH NULL 'NULL' CSV QUOTE '\"' DELIMITER',' ESCAPE '\"' ; "
		self.pg_conn.pgsql_cur.copy_expert(sql_copy,csv_file)
		
	def insert_data(self, table,  insert_data,  my_tables={}):
		"""
			The method is a fallback procedure for when the copy method fails.
			The procedure performs a row by row insert, very slow but capable to skip the rows with problematic data (e.g. enchoding issues).
			
			:param table: the table name, used to get the table's metadata out of my_tables
			:param csv_file: file like object with the table's data stored in CSV format
			:param my_tables: table's metadata dictionary 
		"""
		column_copy=[]
		column_marker=[]
		
		for column in my_tables[table]["columns"]:
			column_copy.append('"'+column["column_name"]+'"')
			column_marker.append('%s')
		sql_head="INSERT INTO "+'"'+self.dest_schema+'"'+"."+'"'+table+'"'+" ("+','.join(column_copy)+") VALUES ("+','.join(column_marker)+");"
		for data_row in insert_data:
			column_values=[]
			for column in my_tables[table]["columns"]:
				column_values.append(data_row[column["column_name"]])
			try:
				self.pg_conn.pgsql_cur.execute(sql_head,column_values)	
			except psycopg2.Error as e:
					self.logger.error("SQLCODE: %s SQLERROR: %s" % (e.pgcode, e.pgerror))
					self.logger.error(self.pg_conn.pgsql_cur.mogrify(sql_head,column_values))
			except ValueError:
				self.logger.warning("character mismatch when inserting the data, trying to cleanup the row data")
				column_values = [str(item).replace("\x00", "") for item in column_values]
				try:	
					self.pg_conn.pgsql_cur.execute(sql_head,column_values)	
				except:
					self.logger.error("error when inserting the row, skipping the row")
				
			except:
				self.logger.error("unexpected error when processing the row")
				self.logger.error(" - > Table: %s" % table)
							
	
	def build_idx_ddl(self):
		""" 
			The method loops over the list l_pkeys and builds a new list with the statements for pkeys 
		"""
		for table_name in self.table_metadata:
			table=self.table_metadata[table_name]
			
			table_name=table["name"]
			indices=table["indices"]
			table_idx=[]
			for index in indices:
				table_timestamp = str(int(time.time()))
				indx=index["index_name"]
				index_columns=index["index_columns"]
				non_unique=index["non_unique"]
				if indx=='PRIMARY':
					pkey_name="pk_"+table_name[0:10]+"_"+table_timestamp+"_"+str(self.idx_sequence)
					pkey_def='ALTER TABLE "'+table_name+'" ADD CONSTRAINT "'+pkey_name+'" PRIMARY KEY ('+index_columns+') ;'
					table_idx.append(pkey_def)
				else:
					if non_unique==0:
						unique_key='UNIQUE'
					else:
						unique_key=''
					index_name='"idx_'+indx[0:10]+table_name[0:10]+"_"+table_timestamp+"_"+str(self.idx_sequence)+'"'
					idx_def='CREATE '+unique_key+' INDEX '+ index_name+' ON "'+table_name+'" ('+index_columns+');'
					table_idx.append(idx_def)
				self.idx_sequence+=1
					
			self.idx_ddl[table_name]=table_idx

	def build_tab_ddl(self):
		""" 
			The method iterates over the list l_tables and builds a new list with the statements for tables
		"""
		if self.table_limit[0] != '*' :
			table_metadata = {}
			for tab in self.table_limit:
				try:
					table_metadata[tab] = self.table_metadata[tab]
				except:
					pass
		else:
			table_metadata = self.table_metadata
		
		for table_name in table_metadata:
			table=self.table_metadata[table_name]
			columns=table["columns"]
			
			ddl_head="CREATE TABLE "+'"'+table["name"]+'" ('
			ddl_tail=");"
			ddl_columns=[]
			ddl_enum=[]
			for column in columns:
				if column["is_nullable"]=="NO":
					col_is_null="NOT NULL"
				else:
					col_is_null="NULL"
				column_type = self.get_data_type(column, table)
				if column_type=="enum":
					enum_type="enum_"+table["name"]+"_"+column["column_name"]
					sql_drop_enum='DROP TYPE IF EXISTS '+enum_type+' CASCADE;'
					sql_create_enum="CREATE TYPE "+enum_type+" AS ENUM "+column["enum_list"]+";"
					ddl_enum.append(sql_drop_enum)
					ddl_enum.append(sql_create_enum)
					column_type=enum_type
				if column_type=="character varying" or column_type=="character":
					column_type=column_type+"("+str(column["character_maximum_length"])+")"
				if column_type=='numeric':
					column_type=column_type+"("+str(column["numeric_precision"])+","+str(column["numeric_scale"])+")"
				if column["extra"]=="auto_increment":
					column_type="bigserial"
				ddl_columns.append('"'+column["column_name"]+'" '+column_type+" "+col_is_null )
			def_columns=str(',').join(ddl_columns)
			self.type_ddl[table["name"]]=ddl_enum
			self.table_ddl[table["name"]]=ddl_head+def_columns+ddl_tail
	

	def get_data_type(self, column, table):
		""" 
			The method determines whether the specified type has to be overridden or not.
			
			:param column: the column dictionary extracted from the information_schema or build in the sql_parser class
			:param table: the table name 
			:return: the postgresql converted column type
			:rtype: string
		"""
		try:
			type_override = self.type_override[column["column_type"]]
			override_to = type_override["override_to"]
			override_tables = type_override["override_tables"]
			
			if override_tables[0] == '*' or table in override_tables:
				column_type = override_to
			else:
				column_type = self.type_dictionary[column["data_type"]]
		except:
			column_type = self.type_dictionary[column["data_type"]]
		return column_type
	
	def get_schema_version(self):
		"""
			The method gets the service schema version querying the view sch_chameleon.v_version.
			The try-except is used in order to get a valid value "base" if the view is missing.
			This happens only if the schema upgrade is performed from very early pg_chamelon's versions.
			
			:return: the catalogg version
			:rtype: string
		"""
		sql_check="""
			SELECT 
				t_version
			FROM 
				sch_chameleon.v_version 
			;
		"""
		try:
			self.pg_conn.pgsql_cur.execute(sql_check)
			value_check=self.pg_conn.pgsql_cur.fetchone()
			cat_version=value_check[0]
		except:
			cat_version='base'
		return cat_version
		
	def upgrade_service_schema(self):
		"""
			The method upgrades the service schema to the latest version using the upgrade files if required.
			
			The method uses the install_script flag to determine whether an upgrade file should be applied.
			The variable cat_version stores the schema version. Each element in the class list cat_sql 
			stores the scripts in the upgrade directory along with the catalogue version associated with the install script.
			
			If the current catalogue version stored in cat_version is equal to the script version the install is skipped but the variable
			install_script is set to true. This way any following install script is executed to upgrade the catalogue to the higher version.
			
			The hardcoded 0.7 version is required because that version introduced the multi source feature.
			As initially the destination schema were not stored in the migration catalogue, the post migration update 
			is required to save this information in the replica catalogue.
		"""
		
		self.logger.info("Upgrading the service schema")
		install_script=False
		cat_version=self.get_schema_version()
			
		for install in self.cat_sql:
				script_ver=install["version"]
				script_schema=install["script"]
				self.logger.info("script schema %s, detected schema version %s - install_script:%s " % (script_ver, cat_version, install_script))
				if install_script==True:
					sql_view="""
					CREATE OR REPLACE VIEW sch_chameleon.v_version 
						AS
							SELECT %s::TEXT t_version
					;"""
					self.logger.info("Installing file version %s" % (script_ver, ))
					file_schema=open(self.sql_dir+script_schema, 'rb')
					sql_schema=file_schema.read()
					file_schema.close()
					self.pg_conn.pgsql_cur.execute(sql_schema)
					self.pg_conn.pgsql_cur.execute(sql_view, (script_ver, ))
					
					if script_ver=='0.7':
						sql_update="""
							UPDATE sch_chameleon.t_sources
							SET
								t_dest_schema=%s 
							WHERE i_id_source=(
												SELECT 
													i_id_source
												FROM
													sch_chameleon.t_sources
												WHERE
													t_source='default'
													AND t_dest_schema='default'
											)
							;
						"""
						self.pg_conn.pgsql_cur.execute(sql_update, (self.dest_schema, ))
				
				
				if script_ver==cat_version and not install_script:
					self.logger.info("enabling install script")
					install_script=True
		
	def check_service_schema(self):
		"""
			The method checks if the sch_chameleon exists
			
			:return: count from information_schema.schemata
			:rtype: integer
		"""
		sql_check="""
			SELECT 
				count(*)
			FROM 
				information_schema.schemata  
			WHERE 
				schema_name='sch_chameleon'
		"""
			
		self.pg_conn.pgsql_cur.execute(sql_check)
		num_schema=self.pg_conn.pgsql_cur.fetchone()
		return num_schema
	
	def create_service_schema(self):
		"""
			The method installs the service replica service schema sch_chameleon.
		"""
		
		num_schema=self.check_service_schema()
		if num_schema[0]==0:
			for install in self.cat_sql:
				script_ver=install["version"]
				script_schema=install["script"]
				if script_ver=='base':
					self.logger.info("Installing service schema %s" % (script_ver, ))
					file_schema=open(self.sql_dir+script_schema, 'rb')
					sql_schema=file_schema.read()
					file_schema.close()
					self.pg_conn.pgsql_cur.execute(sql_schema)
		else:
			self.logger.error("The service schema is already created")
			
	def get_status(self):
		"""
			The metod lists the sources with the running status and the eventual lag 
			
			:return: psycopg2 fetchall results 
			:rtype: psycopg2 tuple
		"""
		sql_status="""
			SELECT
				t_source,
				t_dest_schema,
				enm_status,
				date_trunc('seconds',now())-ts_last_received lag,
				ts_last_received,
				ts_last_received-ts_last_replay,
				ts_last_replay,
				coalesce(t_source_schema,'')
			FROM 
				sch_chameleon.t_sources
			ORDER BY 
				t_source
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_status)
		results = self.pg_conn.pgsql_cur.fetchall()
		return results
		
	def drop_service_schema(self):
		"""
			The method removes the service schema discarding all the replica references.
			The replicated tables are kept in place though.
		"""
		file_schema=open(self.sql_dir+"drop_schema.sql", 'rb')
		sql_schema=file_schema.read()
		file_schema.close()
		self.pg_conn.pgsql_cur.execute(sql_schema)
	
	def save_master_status(self, master_status, cleanup=False):
		"""
			This method saves the master data determining which log table should be used in the next batch.
			
			The method performs also a cleanup for the logged events the cleanup parameter is true.
			
			:param master_status: the master data with the binlogfile and the log position
			:param cleanup: if true cleans the not replayed batches. This is useful when resyncing a replica.
		"""
		next_batch_id=None
		master_data = master_status[0]
		binlog_name = master_data["File"]
		binlog_position = master_data["Position"]
		try:
			event_time = master_data["Time"]
		except:
			event_time = None
		
		sql_master="""
			INSERT INTO sch_chameleon.t_replica_batch
				(
					i_id_source,
					t_binlog_name, 
					i_binlog_position
				)
			VALUES 
				(
					%s,
					%s,
					%s
				)
			RETURNING i_id_batch
			;
		"""
						
		sql_event="""
			UPDATE sch_chameleon.t_sources 
			SET 
				ts_last_received=to_timestamp(%s),
				v_log_table=ARRAY[v_log_table[2],v_log_table[1]]
				
			WHERE 
				i_id_source=%s
			RETURNING 
				v_log_table[1],
				ts_last_received
			; 
		"""
		
		
		
		
		try:
			if cleanup:
				self.logger.info("cleaning not replayed batches for source %s", self.i_id_source)
				sql_cleanup=""" DELETE FROM sch_chameleon.t_replica_batch WHERE i_id_source=%s AND NOT b_replayed; """
				self.pg_conn.pgsql_cur.execute(sql_cleanup, (self.i_id_source, ))
			self.pg_conn.pgsql_cur.execute(sql_master, (self.i_id_source, binlog_name, binlog_position))
			results=self.pg_conn.pgsql_cur.fetchone()
			next_batch_id=results[0]
		except psycopg2.Error as e:
					self.logger.error("SQLCODE: %s SQLERROR: %s" % (e.pgcode, e.pgerror))
					self.logger.error(self.pg_conn.pgsql_cur.mogrify(sql_master, (self.i_id_source, binlog_name, binlog_position)))
		try:
			self.pg_conn.pgsql_cur.execute(sql_event, (event_time, self.i_id_source, ))
			results = self.pg_conn.pgsql_cur.fetchone()
			log_table_name = results[0]
			db_event_time = results[1]
			self.logger.info("Saved master data for source: %s" %(self.source_name, ) )
			self.logger.debug("Binlog file: %s" % (binlog_name, ))
			self.logger.debug("Binlog position:%s" % (binlog_position, ))
			self.logger.debug("Last event: %s" % (db_event_time, ))
			self.logger.debug("Next log table name: %s" % ( log_table_name, ))
		
		
			
		except psycopg2.Error as e:
					self.logger.error("SQLCODE: %s SQLERROR: %s" % (e.pgcode, e.pgerror))
					self.pg_conn.pgsql_cur.mogrify(sql_event, (event_time, self.i_id_source, ))
		
		return next_batch_id
		
	def get_batch_data(self):
		"""
			The method updates the batch status to started for the given source_id and returns the 
			batch informations.
			
			:return: psycopg2 fetchall results without any manipulation
			:rtype: psycopg2 tuple
			
		"""
		sql_batch="""
			WITH t_created AS
				(
					SELECT 
						max(ts_created) AS ts_created
					FROM 
						sch_chameleon.t_replica_batch  
					WHERE 
							NOT b_processed
						AND	NOT b_replayed
						AND	i_id_source=%s
				)
			UPDATE sch_chameleon.t_replica_batch
			SET 
				b_started=True
			FROM 
				t_created
			WHERE
					t_replica_batch.ts_created=t_created.ts_created
				AND	i_id_source=%s
			RETURNING
				i_id_batch,
				t_binlog_name,
				i_binlog_position,
				(SELECT v_log_table[1] from sch_chameleon.t_sources WHERE i_id_source=%s) as v_log_table
				
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_batch, (self.i_id_source, self.i_id_source, self.i_id_source, ))
		return self.pg_conn.pgsql_cur.fetchall()
	
	def insert_batch(self,group_insert):
		"""
			Fallback method for the batch insert. Each row event is processed
			individually and any problematic row is discarded into the table t_discarded_rows.
			The row is encoded in base64 in order to prevent any encoding or type issue.
			
			:param group_insert: the event data built in mysql_engine
		"""
		
		self.logger.debug("starting insert loop")
		for row_data in group_insert:
			global_data = row_data["global_data"]
			event_data = row_data["event_data"]
			event_update = row_data["event_update"]
			log_table = global_data["log_table"]
			event_time = global_data["event_time"]
			sql_insert="""
				INSERT INTO sch_chameleon."""+log_table+"""
					(
						i_id_batch, 
						v_table_name, 
						v_schema_name, 
						enm_binlog_event, 
						t_binlog_name, 
						i_binlog_position, 
						jsb_event_data,
						jsb_event_update,
						i_my_event_time
					)
					VALUES 
						(
							%s,
							%s,
							%s,
							%s,
							%s,
							%s,
							%s,
							%s,
							%s
						)
				;						
			"""
			try:
				self.pg_conn.pgsql_cur.execute(sql_insert,(
						global_data["batch_id"], 
						global_data["table"],  
						global_data["schema"], 
						global_data["action"], 
						global_data["binlog"], 
						global_data["logpos"], 
						json.dumps(event_data, cls=pg_encoder), 
						json.dumps(event_update, cls=pg_encoder), 
						event_time
					)
				)
			except psycopg2.Error as e:
				if e.pgcode == "22P05":
					self.logger.warning("%s - %s. Trying to cleanup the row" % (e.pgcode, e.pgerror))
					event_data = {key: str(value).replace("\x00", "") for key, value in event_data.items()}
					event_update = {key: str(value).replace("\x00", "") for key, value in event_update.items()}
					try:
						self.pg_conn.pgsql_cur.execute(sql_insert,(
								global_data["batch_id"], 
								global_data["table"],  
								global_data["schema"], 
								global_data["action"], 
								global_data["binlog"], 
								global_data["logpos"], 
								json.dumps(event_data, cls=pg_encoder), 
								json.dumps(event_update, cls=pg_encoder), 
								event_time
							)
						)
					except:
						self.logger.error("Cleanup unsuccessful. Saving the discarded row")
						self.save_discarded_row(row_data,global_data["batch_id"])
						
			except:
				self.logger.error("error when storing event data. saving the discarded row")
				self.save_discarded_row(row_data,global_data["batch_id"])
	
	def save_discarded_row(self,row_data,batch_id):
		"""
			The method saves the discarded row in the table t_discarded_row along with the id_batch.
			The row is encoded in base64 as the t_row_data is a text field.
			
			:param row_data: the row data dictionary
			:param batch_id: the id batch where the row belongs
		"""
		str_data = '%s' %(row_data, )
		hex_row = binascii.hexlify(str_data.encode())
		sql_save="""
			INSERT INTO sch_chameleon.t_discarded_rows
				(
					i_id_batch, 
					t_row_data
				)
			VALUES 
				(
					%s,
					%s
				);
		"""
		self.pg_conn.pgsql_cur.execute(sql_save,(batch_id,hex_row))
	
	def write_batch(self, group_insert):
		"""
			Main method for adding the batch data in the log tables. 
			The row data from group_insert are mogrified in CSV format and stored in
			the string like object csv_file.
			
			psycopg2's copy expert is used to store the event data in PostgreSQL.
			
			Should any error occur the procedure fallsback to insert_batch.
			
			:param group_insert: the event data built in mysql_engine
		"""
		csv_file=io.StringIO()
		self.set_application_name("writing batch")
		insert_list=[]
		for row_data in group_insert:
			global_data=row_data["global_data"]
			event_data=row_data["event_data"]
			event_update=row_data["event_update"]
			log_table=global_data["log_table"]
			insert_list.append(self.pg_conn.pgsql_cur.mogrify("%s,%s,%s,%s,%s,%s,%s,%s,%s" ,  (
						global_data["batch_id"], 
						global_data["table"],  
						self.dest_schema, 
						global_data["action"], 
						global_data["binlog"], 
						global_data["logpos"], 
						json.dumps(event_data, cls=pg_encoder), 
						json.dumps(event_update, cls=pg_encoder), 
						global_data["event_time"], 
						
					)
				)
			)
											
		csv_data=b"\n".join(insert_list ).decode()
		csv_file.write(csv_data)
		csv_file.seek(0)
		try:
			
			sql_copy="""
				COPY "sch_chameleon"."""+log_table+""" 
					(
						i_id_batch, 
						v_table_name, 
						v_schema_name, 
						enm_binlog_event, 
						t_binlog_name, 
						i_binlog_position, 
						jsb_event_data,
						jsb_event_update,
						i_my_event_time
					) 
				FROM 
					STDIN 
					WITH NULL 'NULL' 
					CSV QUOTE '''' 
					DELIMITER ',' 
					ESCAPE '''' 
				;
			"""
			self.pg_conn.pgsql_cur.copy_expert(sql_copy,csv_file)
		except psycopg2.Error as e:
			self.logger.error("SQLCODE: %s SQLERROR: %s" % (e.pgcode, e.pgerror))
			self.logger.error(csv_data)
			self.logger.error("fallback to inserts")
			self.insert_batch(group_insert)
		self.set_application_name("idle")
			
		
	def set_batch_processed(self, id_batch):
		"""
			The method updates the flag b_processed and sets the processed timestamp for the given batch id.
			The event ids are aggregated into the table t_batch_events used by the replay function.
			
			:param id_batch: the id batch to set as processed
		"""
		self.logger.debug("updating batch %s to processed" % (id_batch, ))
		sql_update=""" 
			UPDATE sch_chameleon.t_replica_batch
				SET
					b_processed=True,
					ts_processed=now()
			WHERE
				i_id_batch=%s
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_update, (id_batch, ))
		self.logger.debug("collecting events id for batch %s " % (id_batch, ))
		sql_collect_events = """
			INSERT INTO
				sch_chameleon.t_batch_events
				(
					i_id_batch,
					i_id_event
				)
			SELECT
				i_id_batch,
				array_agg(i_id_event)
			FROM
			(
				SELECT 
					i_id_batch,
					i_id_event,
					ts_event_datetime
				FROM 
					sch_chameleon.t_log_replica 
				WHERE i_id_batch=%s
				ORDER BY ts_event_datetime
			) t_event
			GROUP BY
					i_id_batch
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_collect_events, (id_batch, ))
		
	def process_batch(self, replica_batch_size):
		"""
			The method calls the function fn_process_batch with the parameters batch size and the id_source.
			
			The plpgsql function returns true if there are still rows to process. When all rows are replayed 
			the method exits.
			
			:param replica_batch_size: the max rows to process in a single function call. 
			
			
		"""
		batch_loop=True
		sql_process="""SELECT sch_chameleon.fn_process_batch(%s,%s);"""
		self.logger.info("Replaying batch for source %s replay size %s rows" % ( self.source_name, replica_batch_size))
		
		while batch_loop:
			try:
				self.set_application_name("replay batch", "replay")
			except:
				self.pg_conn.connect_replay_db()
				
			self.pg_conn.pgsql_cur_replay.execute(sql_process, (replica_batch_size, self.i_id_source))
			batch_result=self.pg_conn.pgsql_cur_replay.fetchone()
			batch_loop=batch_result[0]
			
			if batch_loop:
				self.logger.info("Still working on batch for source  %s replay size %s rows" % (self.source_name, replica_batch_size ))
			else:
				self.logger.info("Batch replay for source %s is complete" % (self.source_name))
		
		self.set_application_name("cleanup batch", "replay")
		self.logger.debug("Cleanup for replayed batches older than %s for source %s" % (self.batch_retention,  self.source_name))
		sql_cleanup="""
			DELETE FROM 
				sch_chameleon.t_replica_batch
			WHERE
					b_started
				AND b_processed
				AND b_replayed
				AND now()-ts_replayed>%s::interval
				AND i_id_source=%s
			;
		"""
		self.pg_conn.pgsql_cur_replay.execute(sql_cleanup, (self.batch_retention, self.i_id_source ))
		self.set_application_name("idle", "replay")

	def add_foreign_keys(self, source_name, fk_metadata):
		"""
			the method  creates the foreign keys extracted from the mysql catalog
			the keys are created initially as invalid then validated. If an error happens
			is displayed on the log destination
			
			:param source_name: the source name, required to determine the destination schema
			:param fk_metadata: the foreign keys metadata extracted from mysql's information schema
		"""
		fk_list = []
		sql_schema="""
			SELECT
				t_dest_schema 
			FROM
				sch_chameleon.t_sources 
			WHERE
				t_source=%s
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_schema, (source_name, ))
		dschema=self.pg_conn.pgsql_cur.fetchone()
		destination_schema = dschema[0]
		self.logger.info("creating the not validated foreign keys in schema %s" % destination_schema)
		fk_counter = 0
		for foreign_key in fk_metadata:
			table_name = foreign_key["table_name"]
			fk_name = foreign_key["constraint_name"][0:20] + "_" + str(fk_counter)
			fk_cols = foreign_key["fk_cols"]
			referenced_table_name = foreign_key["referenced_table_name"]
			ref_columns = foreign_key["ref_columns"]
			fk_list.append({'fkey_name':fk_name, 'table_name':table_name})
			sql_fkey = ("""ALTER TABLE "%s"."%s" ADD CONSTRAINT "%s" FOREIGN KEY (%s) REFERENCES "%s"."%s" (%s) NOT VALID;""" % 
				(
					destination_schema, 
					table_name, 
					fk_name, 
					fk_cols, 
					destination_schema, 
					referenced_table_name, 
					ref_columns
				)
				)
			fk_counter+=1
			self.logger.debug("creating %s on %s" % (fk_name, table_name))
			try:
				self.pg_conn.pgsql_cur.execute(sql_fkey)
			except psycopg2.Error as e:
					self.logger.error("could not create the foreign key %s on table %s" % (fk_name, table_name))
					self.logger.error("SQLCODE: %s SQLERROR: %s" % (e.pgcode, e.pgerror))
					self.logger.error("STATEMENT: %s " % (self.pg_conn.pgsql_cur.mogrify(sql_fkey)))
					
		self.logger.info("validating the foreign keys in schema %s" % destination_schema)
		for fkey in fk_list:
			sql_validate = 'ALTER TABLE "%s"."%s" VALIDATE CONSTRAINT "%s";' % (destination_schema, fkey["table_name"], fkey["fkey_name"])
			try:
				self.pg_conn.pgsql_cur.execute(sql_validate)
			except psycopg2.Error as e:
					self.logger.error("could not validate the foreign key %s on table %s" % (fkey["table_name"], fkey["fkey_name"]))
					self.logger.error("SQLCODE: %s SQLERROR: %s" % (e.pgcode, e.pgerror))
					self.logger.error("STATEMENT: %s " % (self.pg_conn.pgsql_cur.mogrify(sql_validate)))
		
	def reset_sequences(self, source_name):
		""" 
			the method resets the sequences to the max value available the associated table 
			:param source_name: the source name, required to determine the destination schema
			
		"""
		sql_schema="""
			SELECT
				t_dest_schema 
			FROM
				sch_chameleon.t_sources 
			WHERE
				t_source=%s
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_schema, (source_name, ))
		dschema=self.pg_conn.pgsql_cur.fetchone()
		destination_schema = dschema[0]
		self.logger.info("resetting the sequences in schema %s" % destination_schema)
		sql_gen_reset=""" 
			SELECT 
				format('SELECT setval(%%L::regclass,(select max(%%I) FROM %%I.%%I));',
					replace(replace(column_default,'nextval(''',''),'''::regclass)',''),
					column_name,
					table_schema,
					table_name
				),
				replace(replace(column_default,'nextval(''',''),'''::regclass)','') as  seq_name
			FROM 
					information_schema.columns
			WHERE 
						table_schema=%s
					AND	column_default like 'nextval%%'
		;"""
		self.pg_conn.pgsql_cur.execute(sql_gen_reset, (destination_schema, ))
		results=self.pg_conn.pgsql_cur.fetchall()
		try:
			for statement in results:
				self.logger.info("resetting the sequence  %s" % statement[1])
				self.pg_conn.pgsql_cur.execute(statement[0])
		except psycopg2.Error as e:
					self.logger.error("SQLCODE: %s SQLERROR: %s" % (e.pgcode, e.pgerror))
					self.logger.error(statement)
		except:
			pass
	
	def  generate_default_statements(self, table, column, create_column=None):
		"""
			The method gets the default value associated with the table and column removing the cast.
			:param table: The table name
			:param table: The column name
			:return: the statements for dropping and creating default value on the affected table
			:rtype: dictionary
		"""
		if not create_column:
			create_column = column
		
		regclass = """ "%s"."%s" """ %(self.dest_schema, table)
		sql_def_val = """
			SELECT 
				(
					SELECT 
						split_part(substring(pg_catalog.pg_get_expr(d.adbin, d.adrelid) for 128),'::',1)
					FROM 
						pg_catalog.pg_attrdef d
					WHERE 
							d.adrelid = a.attrelid 
						AND d.adnum = a.attnum 
						AND a.atthasdef
				) as default_value
				FROM 
					pg_catalog.pg_attribute a
				WHERE 
						a.attrelid = %s::regclass 
					AND a.attname=%s 
					AND NOT a.attisdropped
			;

		"""
		self.pg_conn.pgsql_cur.execute(sql_def_val, (regclass, column ))
		default_value = self.pg_conn.pgsql_cur.fetchone()
		if default_value[0]:
			query_drop_default = """ ALTER TABLE  "%s" ALTER COLUMN "%s" DROP DEFAULT;""" % (table, column)
			query_add_default = """ ALTER TABLE  "%s" ALTER COLUMN "%s" SET DEFAULT %s ; """ % (table, create_column, default_value[0])
		else:
			query_drop_default = ""
			query_add_default = ""
				
		return {'drop':query_drop_default, 'create':query_add_default}

	def build_enum_ddl(self, enm_dic):
		"""
			The method builds the enum DDL using the token data. 
			The postgresql system catalog  is queried to determine whether the enum exists and needs to be altered.
			The alter is not written in the replica log table but executed as single statement as PostgreSQL do not allow the alter being part of a multi command
			SQL.
			
			:param enm_dic: a dictionary with the enumeration details
			:return: a dictionary with the pre_alter and post_alter statements (e.g. pre alter create type , post alter drop type)
			:rtype: dictionary
		"""
		#enm_dic = {'table':table_name, 'column':column_name, 'type':column_type, 'enum_list': enum_list}
		enum_name="enum_%s_%s" % (enm_dic['table'], enm_dic['column'])
		
		sql_check_enum = """
			SELECT 
				typ.typcategory,
				typ.typname,
				sch_typ.nspname as typschema,
				CASE 
					WHEN typ.typcategory='E'
					THEN
					(
						SELECT 
							array_agg(enumlabel) 
						FROM 
							pg_enum 
						WHERE 
							enumtypid=typ.oid
					)
				END enum_list
			FROM
				pg_type typ
				INNER JOIN pg_namespace sch_typ
					ON  sch_typ.oid = typ.typnamespace

			WHERE
					sch_typ.nspname=%s
				AND	typ.typname=%s
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_check_enum, (self.dest_schema,  enum_name))
		type_data=self.pg_conn.pgsql_cur.fetchone()
		return_dic = {}
		pre_alter = ""
		post_alter = ""
		column_type = enm_dic["type"]
		self.logger.debug(enm_dic)
		if type_data:
			if type_data[0] == 'E' and enm_dic["type"] == 'enum':
				self.logger.debug('There is already the enum %s, altering the type')
				new_enums = [val.strip() for val in enm_dic["enum_list"] if val.strip() not in type_data[3]]
				sql_add = []
				for enumeration in  new_enums:
					sql_add =  """ALTER TYPE "%s"."%s" ADD VALUE '%s';""" % (type_data[2], enum_name, enumeration) 
					self.pg_conn.pgsql_cur.execute(sql_add)
				column_type = enum_name
			elif type_data[0] != 'E' and enm_dic["type"] == 'enum':
				self.logger.debug('The column will be altered in enum, creating the type')
				pre_alter = "CREATE TYPE \"%s\" AS ENUM (%s);" % (enum_name, enm_dic["enum_elements"])
				column_type = enum_name
			elif type_data[0] == 'E' and enm_dic["type"] != 'enum':
				self.logger.debug('The column is no longer an enum, dropping the type')
				post_alter = "DROP TYPE \"%s\" " % (enum_name)
		elif not type_data and enm_dic["type"] == 'enum':
				self.logger.debug('Creating a new enumeration type %s' % (enum_name))
				pre_alter = "CREATE TYPE \"%s\" AS ENUM (%s);" % (enum_name, enm_dic["enum_elements"])
				column_type = enum_name

		return_dic["column_type"] = column_type
		return_dic["pre_alter"] = pre_alter
		return_dic["post_alter"]  = post_alter
		return return_dic
		
	def build_alter_table(self, token):
		""" 
			The method builds the alter table statement from the token data.
			The function currently supports the following statements.
			DROP TABLE
			ADD COLUMN 
			CHANGE
			MODIFY
			
			The change and modify are potential source of breakage for the replica because of 
			the mysql implicit fallback data types. 
			For better understanding please have a look to 
			
			http://www.cybertec.at/why-favor-postgresql-over-mariadb-mysql/
			
			:param token: A dictionary with the tokenised sql statement
			:return: query the DDL query in the PostgreSQL dialect
			:rtype: string
			
		"""
		alter_cmd = []
		ddl_pre_alter = []
		ddl_post_alter = []
		query_cmd=token["command"]
		table_name=token["name"]
		for alter_dic in token["alter_cmd"]:
			if alter_dic["command"] == 'DROP':
				alter_cmd.append("%(command)s %(name)s CASCADE" % alter_dic)
			elif alter_dic["command"] == 'ADD':
				
				column_type=self.get_data_type(alter_dic, table_name)
				column_name = alter_dic["name"]
				enum_list = str(alter_dic["dimension"]).replace("'", "").split(",")
				enm_dic = {'table':table_name, 'column':column_name, 'type':column_type, 'enum_list': enum_list, 'enum_elements':alter_dic["dimension"]}
				enm_alter = self.build_enum_ddl(enm_dic)
				ddl_pre_alter.append(enm_alter["pre_alter"])
				ddl_post_alter.append(enm_alter["post_alter"])
				column_type= enm_alter["column_type"]
				if 	column_type in ["character varying", "character", 'numeric', 'bit', 'float']:
						column_type=column_type+"("+str(alter_dic["dimension"])+")"
				if alter_dic["default"]:
					default_value = "DEFAULT %s" % alter_dic["default"]
				else:
					default_value=""
				alter_cmd.append("%s \"%s\" %s NULL %s" % (alter_dic["command"], column_name, column_type, default_value))	
			elif alter_dic["command"] == 'CHANGE':
				sql_rename = ""
				sql_type = ""
				old_column=alter_dic["old"]
				new_column=alter_dic["new"]
				column_name = old_column
				enum_list = str(alter_dic["dimension"]).replace("'", "").split(",")
				
				column_type=self.get_data_type(alter_dic, table_name)
				default_sql = self.generate_default_statements(table_name, old_column, new_column)
				enm_dic = {'table':table_name, 'column':column_name, 'type':column_type, 'enum_list': enum_list, 'enum_elements':alter_dic["dimension"]}
				enm_alter = self.build_enum_ddl(enm_dic)

				ddl_pre_alter.append(enm_alter["pre_alter"])
				ddl_pre_alter.append(default_sql["drop"])
				ddl_post_alter.append(enm_alter["post_alter"])
				ddl_post_alter.append(default_sql["create"])
				column_type= enm_alter["column_type"]
				
				if column_type=="character varying" or column_type=="character" or column_type=='numeric' or column_type=='bit' or column_type=='float':
						column_type=column_type+"("+str(alter_dic["dimension"])+")"
				sql_type = """ALTER TABLE "%s" ALTER COLUMN "%s" SET DATA TYPE %s  USING "%s"::%s ;;""" % (table_name, old_column, column_type, old_column, column_type)
				if old_column != new_column:
					sql_rename="""ALTER TABLE  "%s" RENAME COLUMN "%s" TO "%s" ;""" % (table_name, old_column, new_column)
					
				query = ' '.join(ddl_pre_alter)
				query += sql_type+sql_rename
				query += ' '.join(ddl_post_alter)
				return query

			elif alter_dic["command"] == 'MODIFY':
				column_type=self.get_data_type(alter_dic, table_name)
				column_name = alter_dic["name"]
				
				enum_list = str(alter_dic["dimension"]).replace("'", "").split(",")
				default_sql = self.generate_default_statements(table_name, column_name)
				enm_dic = {'table':table_name, 'column':column_name, 'type':column_type, 'enum_list': enum_list, 'enum_elements':alter_dic["dimension"]}
				enm_alter = self.build_enum_ddl(enm_dic)

				ddl_pre_alter.append(enm_alter["pre_alter"])
				ddl_pre_alter.append(default_sql["drop"])
				ddl_post_alter.append(enm_alter["post_alter"])
				ddl_post_alter.append(default_sql["create"])
				column_type= enm_alter["column_type"]
				if column_type=="character varying" or column_type=="character" or column_type=='numeric' or column_type=='bit' or column_type=='float':
						column_type=column_type+"("+str(alter_dic["dimension"])+")"
				query = ' '.join(ddl_pre_alter)
				query +=  """ALTER TABLE "%s" ALTER COLUMN "%s" SET DATA TYPE %s USING "%s"::%s ;""" % (table_name, column_name, column_type, column_name, column_type)
				query += ' '.join(ddl_post_alter)
				return query
		query = ' '.join(ddl_pre_alter)
		query +=  query_cmd + ' '+ table_name+ ' ' +', '.join(alter_cmd)+" ;"
		query += ' '.join(ddl_post_alter)
		return query


	def truncate_tables(self):
		"""
			The method truncate the tables listed in t_index_def. In order to minimise the risk of lock chain
			the truncate is prepended by a set lock_timeout = 10 seconds. If the lock is not acquired in that time
			the procedure fallsback to a delete and vacuum. 
		"""
		sql_clean=""" 
			SELECT DISTINCT
				format('SET lock_timeout=''10s'';TRUNCATE TABLE %%I.%%I CASCADE;',v_schema,v_table) v_truncate,
				format('DELETE FROM %%I.%%I;',v_schema,v_table) v_delete,
				format('VACUUM %%I.%%I;',v_schema,v_table) v_vacuum,
				format('%%I.%%I',v_schema,v_table) as v_tab,
				v_table
			FROM
				sch_chameleon.t_index_def 
			WHERE
				i_id_source=%s
			ORDER BY 
				v_table
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_clean, (self.i_id_source, ))
		tab_clean=self.pg_conn.pgsql_cur.fetchall()
		for stat_clean in tab_clean:
			st_truncate=stat_clean[0]
			st_delete=stat_clean[1]
			st_vacuum=stat_clean[2]
			tab_name=stat_clean[3]
			try:
				self.logger.info("truncating table %s" % (tab_name,))
				self.pg_conn.pgsql_cur.execute(st_truncate)
				
			except:
				self.logger.info("truncate failed, fallback to delete on table %s" % (tab_name,))
				self.pg_conn.pgsql_cur.execute(st_delete)
				self.logger.info("running vacuum on table %s" % (tab_name,))
				self.pg_conn.pgsql_cur.execute(st_vacuum)

	def drop_src_indices(self):
		"""
			The method executes the index drop statements read from the table t_index_def.
			The method is used when resyncing the replica for removing the indices before the bulk load.
		"""
		sql_idx="""
			SELECT 
				t_drop 
			FROM  
				sch_chameleon.t_index_def 
			WHERE 
				i_id_source=%s;
		"""
		self.pg_conn.pgsql_cur.execute(sql_idx, (self.i_id_source, ))
		idx_drop=self.pg_conn.pgsql_cur.fetchall()
		for drop_stat in idx_drop:
			self.pg_conn.pgsql_cur.execute(drop_stat[0])
			
	def create_src_indices(self):
		"""
			The method executes the index DDL read from the table t_index_def.
			The method is used when resyncing the replica for recreating the indices after the bulk load.
		"""
		sql_idx="""
			SELECT 
				t_create 
			FROM  
				sch_chameleon.t_index_def 
			WHERE 
				i_id_source=%s;
		"""
		self.pg_conn.pgsql_cur.execute(sql_idx, (self.i_id_source, ))
		idx_create=self.pg_conn.pgsql_cur.fetchall()
		for create_stat in idx_create:
			self.pg_conn.pgsql_cur.execute(create_stat[0])
		

	def get_index_def(self):
		"""
			The method inserts in the table t_index_def the create and drop statements for the tables affected by 
			the resync replica.
		"""
		table_limit = ''
		if self.table_limit[0] != '*':
			table_limit = self.pg_conn.pgsql_cur.mogrify("""WHERE table_name IN  (SELECT unnest(%s))""",(self.table_limit, )).decode()
		
		sql_get_idx=""" 
			DELETE FROM sch_chameleon.t_index_def WHERE i_id_source=%s;
			INSERT INTO sch_chameleon.t_index_def
				(
					i_id_source,
					v_schema,
					v_table,
					v_index,
					t_create,
					t_drop
				)
			SELECT 
				i_id_source,
				schema_name,
				table_name,
				index_name,
				CASE
					WHEN indisprimary
					THEN
						format('ALTER TABLE %%I.%%I ADD CONSTRAINT %%I %%s',
							schema_name,
							table_name,
							index_name,
							pg_get_constraintdef(const_id)
						)
						
					ELSE
						pg_get_indexdef(index_id)    
				END AS t_create,
				CASE
					WHEN indisprimary
					THEN
						format('ALTER TABLE %%I.%%I DROP CONSTRAINT %%I',
							schema_name,
							table_name,
							index_name
							
						)
						
					ELSE
						format('DROP INDEX %%I.%%I',
							schema_name,
							index_name
							
						)
				END AS  t_drop
				
			FROM

			(
			SELECT 
				tab.relname AS table_name,
				indx.relname AS index_name,
				idx.indexrelid index_id,
				indisprimary,
				sch.nspname schema_name,
				src.i_id_source,
				cns.oid as const_id
				
			FROM
				pg_index idx
				INNER JOIN pg_class indx
				ON
					idx.indexrelid=indx.oid
				INNER JOIN pg_class tab
				INNER JOIN pg_namespace sch
				ON 
					tab.relnamespace=sch.oid
				
				ON
					idx.indrelid=tab.oid
				INNER JOIN sch_chameleon.t_sources src
				ON sch.nspname=src.t_dest_schema
				LEFT OUTER JOIN pg_constraint cns
				ON 
						indx.relname=cns.conname
					AND cns.connamespace=sch.oid
				
			WHERE
				sch.nspname=%s
			) idx
			
		""" + table_limit
		
		self.pg_conn.pgsql_cur.execute(sql_get_idx, (self.i_id_source,  self.dest_schema, ))

	def drop_primary_key(self, token):
		"""
			The method drops the primary key for the table.
			As tables without primary key cannot be replicated the method calls unregister_table
			to remove the table from the replica set.
			The drop constraint statement is not built from the token but generated from the information_schema.
			
			
			:param token: the tokenised query for drop primary key
		"""
		self.logger.info("dropping primary key for table %s" % (token["name"],))
		sql_gen="""
			SELECT  DISTINCT
				format('ALTER TABLE %%I.%%I DROP CONSTRAINT %%I;',
				table_schema,
				table_name,
				constraint_name
				)
			FROM 
				information_schema.key_column_usage 
			WHERE 
					table_schema=%s 
				AND table_name=%s
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_gen, (self.dest_schema, token["name"]))
		value_check=self.pg_conn.pgsql_cur.fetchone()
		if value_check:
			sql_drop=value_check[0]
			self.pg_conn.pgsql_cur.execute(sql_drop)
			self.unregister_table(token["name"])
		

	def gen_query(self, token):
		""" 
			The method builds the DDL using the tokenised SQL stored in token.
			The supported commands are 
			RENAME TABLE
			DROP TABLE
			TRUNCATE
			CREATE TABLE
			ALTER TABLE
			DROP PRIMARY KEY
			
			:param token: A dictionary with the tokenised sql statement
			:return: query the DDL query in the PostgreSQL dialect
			:rtype: string
			
		"""
		query=""
		if token["command"] =="RENAME TABLE":
			query = """ALTER TABLE "%s" RENAME TO "%s" """ % (token["name"], token["new_name"])	
			try:
				self.table_metadata[token["new_name"]]
				self.store_table(token["new_name"])
			except KeyError:
				try:
					self.table_metadata[token["new_name"]] = self.table_metadata[token["name"]]
					self.store_table(token["new_name"])
				except KeyError:
					query = ""
			
		elif token["command"] =="DROP TABLE":
			query=" %(command)s IF EXISTS \"%(name)s\";" % token
		elif token["command"] =="TRUNCATE":
			query=" %(command)s TABLE \"%(name)s\" CASCADE;" % token
		elif token["command"] =="CREATE TABLE":
			table_metadata={}
			table_metadata["columns"]=token["columns"]
			table_metadata["name"]=token["name"]
			table_metadata["indices"]=token["indices"]
			self.table_metadata={}
			self.table_metadata[token["name"]]=table_metadata
			self.build_tab_ddl()
			self.build_idx_ddl()
			query_type=' '.join(self.type_ddl[token["name"]])
			query_table=self.table_ddl[token["name"]]
			query_idx=' '.join(self.idx_ddl[token["name"]])
			query=query_type+query_table+query_idx
			self.store_table(token["name"])
		elif token["command"] == "ALTER TABLE":
			query=self.build_alter_table(token)
		elif token["command"] == "DROP PRIMARY KEY":
			self.drop_primary_key(token)
		return query 


	def write_ddl(self, token, query_data):
		"""
			The method writes the DDL built from the tokenised sql into PostgreSQL.
			
			:param token: the tokenised query
			:param query_data: query's metadata (schema,binlog, etc.)
		"""
		sql_path=""" SET search_path="%s";""" % self.dest_schema
		pg_ddl=sql_path+self.gen_query(token)
		log_table=query_data["log_table"]
		insert_vals=(	query_data["batch_id"], 
								token["name"],  
								query_data["schema"], 
								query_data["binlog"], 
								query_data["logpos"], 
								pg_ddl
							)
		sql_insert="""
			INSERT INTO sch_chameleon."""+log_table+"""
				(
					i_id_batch, 
					v_table_name, 
					v_schema_name, 
					enm_binlog_event, 
					t_binlog_name, 
					i_binlog_position, 
					t_query
				)
			VALUES
				(
					%s,
					%s,
					%s,
					'ddl',
					%s,
					%s,
					%s
				)
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_insert, insert_vals)
		
		
	def check_reindex(self):
		"""
			the function checks if there is any reindex running and holds for  the given number of seconds 
		"""
		sql_check="""
			SELECT 
				count(*) 
			FROM 
				pg_stat_activity 
			WHERE 
					datname=current_database() 
				AND	application_name = ANY(%s) ;
		"""
		while True:
			self.pg_conn.pgsql_cur.execute(sql_check, (self.reindex_app_names, ))
			reindex_tup = self.pg_conn.pgsql_cur.fetchone()
			reindex_cnt = reindex_tup[0]
			if reindex_cnt == 0:
				break;
			self.logger.info("reindex detected, sleeping %s second(s)" % (self.sleep_on_reindex,))
			time.sleep(self.sleep_on_reindex)
	
	def set_consistent_table(self, table):
		"""
			The method set to NULL the  binlog name and position for the given table.
			When the table is marked consistent the read replica loop reads and saves the table's row images.
			
			:param table: the table name
		"""
		sql_set = """
			UPDATE sch_chameleon.t_replica_tables
				SET 
					t_binlog_name = NULL,
					i_binlog_position = NULL
			WHERE
					i_id_source = %s
				AND	v_table_name = %s
				AND	v_schema_name = %s
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_set, (self.i_id_source, table, self.dest_schema))
		
	
	def get_inconsistent_tables(self):
		"""
			The method collects the tables in not consistent state.
			The informations are stored in a dictionary which key is the table's name.
			The dictionary is used in the read replica loop to determine wheter the table's modifications
			should be ignored because in not consistent state.
			
			:return: a dictionary with the tables in inconsistent state and their snapshot coordinates.
			:rtype: dictionary
		"""
		sql_get = """
			SELECT
				v_schema_name,				
				v_table_name,
				t_binlog_name,
				i_binlog_position
			FROM
				sch_chameleon.t_replica_tables
			WHERE
				t_binlog_name IS NOT NULL
				AND i_binlog_position IS NOT NULL
				AND i_id_source = %s
		;
		"""
		inc_dic = {}
		self.pg_conn.pgsql_cur.execute(sql_get, (self.i_id_source, ))
		inc_results = self.pg_conn.pgsql_cur.fetchall()
		for table  in inc_results:
			tab_dic = {}
			tab_dic["schema"]  = table[0]
			tab_dic["table"]  = table[1]
			tab_dic["log_seq"]  = int(table[2].split('.')[1])
			tab_dic["log_pos"]  = int(table[3])
			inc_dic[table[1]] = tab_dic
		return inc_dic
		
		
	def delete_table_events(self):
		"""
			The method removes the events from the log table for specific table and source. 
			Is used to cleanup any residual event for a a synced table in the replica_engine's sync_table method.
		"""
		sql_clean = """
			DELETE FROM sch_chameleon.t_log_replica
			WHERE 
				i_id_event IN (
							SELECT 
								log.i_id_event
							FROM
								sch_chameleon.t_replica_batch bat
								INNER JOIN sch_chameleon.t_log_replica log
									ON  log.i_id_batch=bat.i_id_batch
							WHERE
									log.v_table_name=ANY(%s)
								AND 	bat.i_id_source=%s
						)
			;
		"""
		self.pg_conn.pgsql_cur.execute(sql_clean, (self.table_limit, self.i_id_source, ))
		



#!/usr/bin/env python
from pkg_resources import get_distribution
__version__ = get_distribution('pg_chameleon').version
import argparse
from pg_chameleon import replica_engine
from pg_chameleon import config_dir
configdir = config_dir()
configdir.set_config()

commands = [
	'create_schema',
	'init_replica',
	'start_replica',
	'upgrade_schema',
	'drop_schema', 
	'list_config', 
	'add_source', 
	'drop_source', 
	'stop_replica', 
	'disable_replica', 
	'enable_replica', 
	'sync_tables', 
	'show_status' , 
	'detach_replica', 
	'set_config'
	]
command_help = 'Available commands, ' + ','.join(commands)
table_help =  """Specifies the table's name to sync. It's possible to specify multiple table names separated by comma. If the parameter is omitted all tables will be syncronised."""
config_help =  """Specifies the configuration to use. The configuration shall be specified without extension (e.g. --config foo) and the file foo.yaml should be in ~/.pg_chameleon/config/"""
debug_help = """Enables the debug mode.The log output is stdout and in debug verbosity. The lock file creation on error is skipped as well."""
nolock_help = """Do not create the lock file on error, allowing the replica to be restarted immediately. """
thread_help = """  When specified the replica process runs the read and replay in two threads."""

parser = argparse.ArgumentParser(description='Command line for pg_chameleon.',  add_help=True)
parser.add_argument('command', metavar='command', type=str, help=command_help)
parser.add_argument('--config', metavar='config', type=str,  default='default',  required=False, help=config_help)
parser.add_argument('--table', metavar='table', type=str,  default='*',  required=False, help=table_help)
parser.add_argument('--debug',  default=False,  required=False, help=nolock_help,  action='store_true')
parser.add_argument('--nolock',  default=False,  required=False, help=debug_help,  action='store_true')
parser.add_argument('--version', action='version',version='pg_chameleon {version}'.format(version=__version__))
parser.add_argument('--thread',  default=False,  required=False, help=thread_help,  action='store_true')
args = parser.parse_args()

if args.command == commands[14]:
		configdir.set_config()
elif args.command in commands:
	
	replica = replica_engine(args.config, args.debug, args.nolock)
	if args.command == commands[0]:
		replica.create_service_schema()
	elif args.command == commands[1]:
		replica.init_replica()
	elif args.command == commands[2]:
		if args.thread:
			replica.run_replica_thread()
		else:
			replica.run_replica()
	elif args.command == commands[3]:
		replica.upgrade_service_schema()
	elif args.command == commands[4]:
		replica.drop_service_schema()
	elif args.command == commands[5]:
		replica.list_config()
	elif args.command == commands[6]:
		replica.add_source()
	elif args.command == commands[7]:
		replica.drop_source()
	elif args.command == commands[8]:
		replica.stop_replica()
	elif args.command == commands[9]:
		replica.stop_replica(allow_restart=False)
	elif args.command == commands[10]:
		replica.enable_replica()
	elif args.command == commands[11]:
		replica.sync_tables(args.table)
	elif args.command == commands[12]:
		replica.show_status()
	elif args.command == commands[13]:
		replica.detach_replica()
	



import io
import pymysql
import codecs
import binascii
from pymysqlreplication import BinLogStreamReader
from pymysqlreplication.event import QueryEvent
from pymysqlreplication.row_event import DeleteRowsEvent,UpdateRowsEvent,WriteRowsEvent
from pymysqlreplication.event import RotateEvent
from pg_chameleon import sql_token
from os import remove

class mysql_connection(object):
	def __init__(self, global_config):
		self.global_conf = global_config
		self.my_server_id = self.global_conf.my_server_id
		self.mysql_conn = self.global_conf.mysql_conn
		self.my_database = self.global_conf.my_database
		self.my_charset = self.global_conf.my_charset
		self.tables_limit = self.global_conf.tables_limit
		self.replica_batch_size = self.global_conf.replica_batch_size
		self.copy_mode = self.global_conf.copy_mode
		self.my_connection = None
		self.my_cursor = None
		self.my_cursor_fallback = None
		
	def connect_db_ubf(self):
		""" Connects to the database and creates an unbuffered cursor """
		self.my_connection_ubf=pymysql.connect(
			host=self.mysql_conn["host"],
			user=self.mysql_conn["user"],
			password=self.mysql_conn["passwd"],
			db=self.my_database,
			charset=self.my_charset,
			cursorclass=pymysql.cursors.SSCursor
		)
		self.my_cursor_ubf=self.my_connection_ubf.cursor()

		
	
	def connect_db(self):
		""" Connects to the database and creates a dictionary cursor """
		self.my_connection=pymysql.connect(
			host=self.mysql_conn["host"],
			user=self.mysql_conn["user"],
			password=self.mysql_conn["passwd"],
			db=self.my_database,
			charset=self.my_charset,
			cursorclass=pymysql.cursors.DictCursor
		)
		self.my_cursor=self.my_connection.cursor()
		self.my_cursor_fallback=self.my_connection.cursor()
	
	def disconnect_db(self):
		"""Disconnects the dictionary connection"""
		try:
			self.my_connection.close()
		except:
			pass
	def disconnect_db_ubf(self):
		"""Disconnects the unbuffered connection"""
		try:
			self.my_connection_ubf.close()
		except:
			pass
		
class mysql_engine(object):
	def __init__(self, global_config, logger):
		"""
			Class constructor
		"""
		self.hexify = global_config.hexify
		self.logger = logger
		self.out_dir = global_config.out_dir
		self.my_tables = {}
		self.table_file = {}
		self.mysql_con = mysql_connection(global_config)
		self.mysql_con.connect_db()
		self.get_table_metadata()
		self.my_streamer = None
		self.replica_batch_size = self.mysql_con.replica_batch_size
		self.master_status = []
		self.id_batch = None
		self.sql_token = sql_token()
		self.pause_on_reindex = global_config.pause_on_reindex
		self.stat_skip = ['BEGIN', 'COMMIT']
		self.tables_limit = global_config.tables_limit
		self.my_schema = global_config.my_database
	
	def read_replica(self, batch_data, pg_engine):
		"""
		Stream the replica using the batch data. This method evaluates the different events streamed from MySQL 
		and manages them accordingly. The BinLogStreamReader function is called with the only_event parameter which
		restricts the event type received by the streamer.
		The events managed are the following.
		RotateEvent which happens whether mysql restarts or the binary log file changes.
		QueryEvent which happens when a new row image comes in (BEGIN statement) or a DDL is executed.
		The BEGIN is always skipped. The DDL is parsed using the sql_token class. 
		[Write,Update,Delete]RowEvents are the row images pulled from the mysql replica.
		
		The RotateEvent and the QueryEvent cause the batch to be closed.
		
		The for loop reads the row events, builds the dictionary carrying informations like the destination schema,
		the 	binlog coordinates and store them into the group_insert list.
		When the number of events exceeds the replica_batch_size the group_insert is written into PostgreSQL.
		The batch is not closed in that case and the method exits only if there are no more rows available in the stream.
		Therefore the replica_batch_size is just the maximum size of the single insert and the size of replayed batch on PostgreSQL.
		The binlog switch or a captured DDL determines whether a batch is closed and processed.
		
		The update row event stores in a separate key event_update the row image before the update. This is required
		to allow updates where the primary key is updated as well.
		
		Each row event is scanned for data types requiring conversion to hex string.
		
		:param batch_data: The list with the master's batch data.
		:param pg_engine: The postgresql engine object required for writing the rows in the log tables
		"""
		table_type_map = self.get_table_type_map()	
		schema_name = pg_engine.dest_schema
		inc_tables = pg_engine.get_inconsistent_tables()
		close_batch = False
		master_data = {}
		group_insert = []
		id_batch = batch_data[0][0]
		log_file = batch_data[0][1]
		log_position = batch_data[0][2]
		log_table = batch_data[0][3]
		my_stream = BinLogStreamReader(
			connection_settings = self.mysql_con.mysql_conn, 
			server_id = self.mysql_con.my_server_id, 
			only_events = [RotateEvent, DeleteRowsEvent, WriteRowsEvent, UpdateRowsEvent, QueryEvent], 
			log_file = log_file, 
			log_pos = log_position, 
			resume_stream = True, 
			only_schemas = [self.mysql_con.my_database], 
			only_tables = self.tables_limit
		)
		self.logger.debug("log_file %s, log_position %s. id_batch: %s " % (log_file, log_position, id_batch))
		for binlogevent in my_stream:
			if isinstance(binlogevent, RotateEvent):
				
				event_time = binlogevent.timestamp
				binlogfile = binlogevent.next_binlog
				position = binlogevent.position
				self.logger.debug("ROTATE EVENT - binlogfile %s, position %s. " % (binlogfile, position))
				if log_file != binlogfile:
					close_batch = True
				if close_batch:
					if log_file!=binlogfile:
						master_data["File"]=binlogfile
						master_data["Position"]=position
						master_data["Time"]=event_time
					if len(group_insert)>0:
						pg_engine.write_batch(group_insert)
						group_insert=[]
					my_stream.close()
					return [master_data, close_batch]
			elif isinstance(binlogevent, QueryEvent):
				event_time = binlogevent.timestamp
				try:
					query_schema = binlogevent.schema.decode()
				except:
					query_schema = binlogevent.schema
				if binlogevent.query.strip().upper() not in self.stat_skip and query_schema == self.my_schema: 
					log_position = binlogevent.packet.log_pos
					master_data["File"] = binlogfile
					master_data["Position"] = log_position
					master_data["Time"] = event_time
					if len(group_insert)>0:
						pg_engine.write_batch(group_insert)
						group_insert=[]
					self.logger.info("QUERY EVENT - binlogfile %s, position %s.\n--------\n%s\n-------- " % (binlogfile, log_position, binlogevent.query))
					self.sql_token.parse_sql(binlogevent.query)
					for token in self.sql_token.tokenised:
						write_ddl = True
						table_name = token["name"] 
						if table_name in inc_tables:
							write_ddl = False
							log_seq = int(log_file.split('.')[1])
							log_pos = int(log_position)
							table_dic = inc_tables[table_name]
							if log_seq > table_dic["log_seq"]:
								write_ddl = True
							elif log_seq == table_dic["log_seq"] and log_pos >= table_dic["log_pos"]:
								write_ddl = True
							if write_ddl:
								self.logger.info("CONSISTENT POINT FOR TABLE %s REACHED  - binlogfile %s, position %s" % (table_name, binlogfile, log_position))
								pg_engine.set_consistent_table(table_name)
								inc_tables = pg_engine.get_inconsistent_tables()
						if write_ddl:
							self.get_table_metadata()
							pg_engine.table_metadata = self.my_tables
							event_time = binlogevent.timestamp
							self.logger.debug("TOKEN: %s" % (token))
							if len(token)>0:
								query_data={
									"binlog":log_file, 
									"logpos":log_position, 
									"schema": schema_name, 
									"batch_id":id_batch, 
									"log_table":log_table
								}
								pg_engine.write_ddl(token, query_data)
								close_batch=True
							
						
					self.sql_token.reset_lists()
					if close_batch:
						my_stream.close()
						return [master_data, close_batch]
			else:
				size_insert=0
				for row in binlogevent.rows:
					add_row = True
					log_file=binlogfile
					log_position=binlogevent.packet.log_pos
					table_name=binlogevent.table
					event_time=binlogevent.timestamp
					if table_name in inc_tables:
						table_consistent = False
						log_seq = int(log_file.split('.')[1])
						log_pos = int(log_position)
						table_dic = inc_tables[table_name]
						if log_seq > table_dic["log_seq"]:
							table_consistent = True
						elif log_seq == table_dic["log_seq"] and log_pos >= table_dic["log_pos"]:
							table_consistent = True
							
						if table_consistent:
							add_row = True
							self.logger.debug("CONSISTENT POINT FOR TABLE %s REACHED  - binlogfile %s, position %s" % (table_name, binlogfile, log_position))
							pg_engine.set_consistent_table(table_name)
							inc_tables = pg_engine.get_inconsistent_tables()
						else:
							add_row = False
							
					column_map=table_type_map[table_name]
					global_data={
										"binlog":log_file, 
										"logpos":log_position, 
										"schema": schema_name, 
										"table": table_name, 
										"batch_id":id_batch, 
										"log_table":log_table, 
										"event_time":event_time
									}
					event_data={}
					event_update={}
					if add_row:
						event_insert = {}
						if isinstance(binlogevent, DeleteRowsEvent):
							global_data["action"] = "delete"
							event_data=row["values"]
						elif isinstance(binlogevent, UpdateRowsEvent):
							global_data["action"] = "update"
							event_data=row["after_values"]
							event_update=row["before_values"]
						elif isinstance(binlogevent, WriteRowsEvent):
							global_data["action"] = "insert"
							event_data=row["values"]
						for column_name in event_data:
							column_type=column_map[column_name]
							if column_type in self.hexify and event_data[column_name]:
								event_data[column_name]=binascii.hexlify(event_data[column_name]).decode()
							elif column_type in self.hexify and isinstance(event_data[column_name], bytes):
								event_data[column_name] = ''
						for column_name in event_update:
							column_type=column_map[column_name]
							if column_type in self.hexify and event_update[column_name]:
								event_update[column_name]=binascii.hexlify(event_update[column_name]).decode()
							elif column_type in self.hexify and isinstance(event_update[column_name], bytes):
								event_update[column_name] = ''
						event_insert={"global_data":global_data,"event_data":event_data,  "event_update":event_update}
						size_insert += len(str(event_insert))
						group_insert.append(event_insert)
						
					master_data["File"]=log_file
					master_data["Position"]=log_position
					master_data["Time"]=event_time
					
					if len(group_insert)>=self.replica_batch_size:
						self.logger.debug("Max rows per batch reached. Writing %s. rows. Size in bytes: %s " % (len(group_insert), size_insert))
						self.logger.debug("Master coordinates: %s" % (master_data, ))
						pg_engine.write_batch(group_insert)
						size_insert=0
						group_insert=[]
						close_batch=True
						
						
						
		my_stream.close()
		if len(group_insert)>0:
			self.logger.debug("writing the last %s events" % (len(group_insert), ))
			pg_engine.write_batch(group_insert)
			close_batch=True
		
		return [master_data, close_batch]
	
	def check_mysql_config(self):
		"""
			The method check if the mysql configuration is compatible with the replica requirements.
			If all the configuration requirements are met then the return value is True.
			Otherwise is false.
			The parameters checked are
			log_bin - ON if the binary log is enabled
			binlog_format - must be ROW , otherwise the replica won't get the data
			binlog_row_image - must be FULL, otherwise the row image will be incomplete
			
			:return: true if all the requirements are met, false if not
			:rtype: boolean
		"""
		sql_log_bin = """SHOW GLOBAL VARIABLES LIKE 'log_bin';"""
		self.mysql_con.my_cursor.execute(sql_log_bin)
		variable_check = self.mysql_con.my_cursor.fetchone()
		log_bin = variable_check["Value"]
		
		sql_log_bin = """SHOW GLOBAL VARIABLES LIKE 'binlog_format';"""
		self.mysql_con.my_cursor.execute(sql_log_bin)
		variable_check = self.mysql_con.my_cursor.fetchone()
		binlog_format = variable_check["Value"]
		
		sql_log_bin = """SHOW GLOBAL VARIABLES LIKE 'binlog_row_image';"""
		self.mysql_con.my_cursor.execute(sql_log_bin)
		variable_check = self.mysql_con.my_cursor.fetchone()
		binlog_row_image = variable_check["Value"]
		if log_bin.upper() == 'ON' and binlog_format.upper() == 'ROW' and binlog_row_image.upper() == 'FULL':
			replica_possible = True
		else:
			replica_possible = False
		return replica_possible
		
	def run_replica(self, pg_engine):
		"""
		Run a MySQL replica read attempt and stores the data in postgres if found. 
		The method first checks if there is a reindex in progress using pg_engine.check_reindex.
		The gets the batch data from PostgreSQL. If the batch data is not empty the method read_replica is executed.
		When the method exits the replica_data list is decomposed in the master_data (log name, position and last event's timestamp).
		If the flag close_batch is set the master status is saved in PostgreSQL, and the new batch id is saved in the mysql_engine class variable id_batch.
		This variable is used to determine whether the old batch should be closed or not. If the variable is not empty then the previous batch
		is closed updating the processed flag on the replica catalogue.
		Before the exit the method calls the pg_engine.process_batch to replay the changes in PostgreSQL
		
		:param pg_engine: The postgresql engine object required for storing the master coordinates and replaying the batches
		"""
		#self.mysql_con.connect_db()
		#pg_engine.pg_conn.connect_db()
		if self.pause_on_reindex:
			pg_engine.check_reindex()
		batch_data=pg_engine.get_batch_data()
		self.logger.debug('batch data: %s' % (batch_data, ))
		if len(batch_data)>0:
			id_batch=batch_data[0][0]
			replica_data=self.read_replica(batch_data, pg_engine)
			master_data=replica_data[0]
			close_batch=replica_data[1]
			if close_batch:
				self.master_status=[]
				self.master_status.append(master_data)
				self.logger.debug("trying to save the master data...")
				next_id_batch=pg_engine.save_master_status(self.master_status)
				if next_id_batch:
					self.logger.debug("new batch created, saving id_batch %s in class variable" % (id_batch))
					self.id_batch=id_batch
				else:
					self.logger.debug("batch not saved. using old id_batch %s" % (self.id_batch))
				if self.id_batch:
					self.logger.debug("updating processed flag for id_batch %s", (id_batch))
					pg_engine.set_batch_processed(id_batch)
					self.id_batch=None


	def get_table_type_map(self):
		"""
			The method builds a dictionary composed by the table name and the column/type mappings.
			The dictionary is used in read_replica to determine whether a field requires hex encoding.
		"""
		table_type_map={}
		self.logger.debug("collecting table type map")
		sql_tables="""	
			SELECT 
				table_schema,
				table_name
			FROM 
				information_schema.TABLES 
			WHERE 
					table_type='BASE TABLE' 
				AND	table_schema=%s
			;
		"""
		self.mysql_con.my_cursor.execute(sql_tables, (self.mysql_con.my_database))
		table_list=self.mysql_con.my_cursor.fetchall()
		for table in table_list:
			column_type={}
			sql_columns="""
				SELECT 
					column_name,
					data_type
				FROM 
					information_schema.COLUMNS 
				WHERE 
						table_schema=%s
					AND table_name=%s
				ORDER BY 
					ordinal_position
				;
			"""
			self.mysql_con.my_cursor.execute(sql_columns, (self.mysql_con.my_database, table["table_name"]))
			column_data=self.mysql_con.my_cursor.fetchall()
			for column in column_data:
				column_type[column["column_name"]]=column["data_type"]
			table_type_map[table["table_name"]]=column_type
		return table_type_map
		
			
		
	def get_column_metadata(self, table):
		"""
			The method extracts the columns metadata for a specific table.
			The select builds also the field list formatted for the CSV copy or a single insert copy.
			The data types included in hexlify are hex encoded on the fly.
			
			:param table: The table name.
			
			'datetime':'timestamp without time zone',
			'date':'date',
			'timestamp':'timestamp without time zone',
			
		"""
		sql_columns="""
			SELECT 
				column_name,
				column_default,
				ordinal_position,
				data_type,
				column_type,
				character_maximum_length,
				extra,
				column_key,
				is_nullable,
				numeric_precision,
				numeric_scale,
				CASE 
					WHEN data_type="enum"
				THEN	
					SUBSTRING(COLUMN_TYPE,5)
				END AS enum_list,
				CASE
					WHEN 
						data_type IN ('"""+"','".join(self.hexify)+"""')
					THEN
						concat('hex(',column_name,')')
					WHEN 
						data_type IN ('bit')
					THEN
						concat('cast(`',column_name,'` AS unsigned)')
					WHEN 
						data_type IN ('datetime','timestamp','date')
					THEN
						concat('nullif(`',column_name,'`,"0000-00-00 00:00:00")')

				ELSE
					concat('cast(`',column_name,'` AS char CHARACTER SET """+ self.mysql_con.my_charset +""")')
				END
				AS column_csv,
				CASE
					WHEN 
						data_type IN ('"""+"','".join(self.hexify)+"""')
					THEN
						concat('hex(',column_name,') AS','`',column_name,'`')
					WHEN 
						data_type IN ('bit')
					THEN
						concat('cast(`',column_name,'` AS unsigned) AS','`',column_name,'`')
					WHEN 
						data_type IN ('datetime','timestamp','date')
					THEN
						concat('nullif(`',column_name,'`,"0000-00-00 00:00:00") AS `',column_name,'`')
					
				ELSE
					concat('cast(`',column_name,'` AS char CHARACTER SET """+ self.mysql_con.my_charset +""") AS','`',column_name,'`')
					
				END
				AS column_select
			FROM 
				information_schema.COLUMNS 
			WHERE 
				table_schema=%s
				AND 	table_name=%s
			ORDER BY 
				ordinal_position
			;
		"""
		self.mysql_con.my_cursor.execute(sql_columns, (self.mysql_con.my_database, table))
		column_data=self.mysql_con.my_cursor.fetchall()
		return column_data


	def get_index_metadata(self, table):
		"""
			The method extracts the index metadata for a specific table.
			The select searches only for the BTREE indices using the information_schema.statistics table.
	
			:param table: The table name.
		"""
		sql_index="""
			SELECT 
				index_name,
				non_unique,
				GROUP_CONCAT(concat('"',column_name,'"') ORDER BY seq_in_index) as index_columns
			FROM
				information_schema.statistics
			WHERE
					table_schema=%s
				AND 	table_name=%s
				AND	index_type = 'BTREE'
			GROUP BY 
				table_name,
				non_unique,
				index_name
			;
		"""
		self.mysql_con.my_cursor.execute(sql_index, (self.mysql_con.my_database, table))
		index_data=self.mysql_con.my_cursor.fetchall()
		return index_data
	
	def get_fk_metadata(self):
		"""
			The method collects the foreign key metadata for the detach replica process.
			Currently doesn't get the ON UPDATE/ON DELETE triggers
		"""
		self.logger.debug("getting foreign keys metadata")
		sql_fkeys = """ 
			SELECT 
				table_name,
				constraint_name,
				referenced_table_name,
				GROUP_CONCAT(concat('"',column_name,'"') ORDER BY POSITION_IN_UNIQUE_CONSTRAINT) as fk_cols,
				GROUP_CONCAT(concat('"',REFERENCED_COLUMN_NAME,'"') ORDER BY POSITION_IN_UNIQUE_CONSTRAINT) as ref_columns
			FROM 
				information_schema.key_column_usage 
			WHERE 
					table_schema=%s 
				AND referenced_table_name IS NOT NULL
				AND referenced_table_schema=%s
				
			GROUP BY 
				table_name,
				constraint_name,
				referenced_table_name
				
			ORDER BY 
				table_name
			;
		"""
		self.mysql_con.my_cursor.execute(sql_fkeys, (self.mysql_con.my_database, self.mysql_con.my_database))
		fkey_list=self.mysql_con.my_cursor.fetchall()
		return fkey_list
		
	def get_table_metadata(self):
		"""
			the metod collects the metadata for all the tables in the mysql schema specified
			in self.my_database using get_column_metadata and get_index_metadata.
			If there are tables in tables_limit the variable table_include is set in order to limit the results to the tables_limit only.
			The informations are stored in a dictionary.
			The key column stores the column metadata and indices stores the index metadata.
			The key name stores the table name.
			The dictionary is then saved in the class dictionary self.my_tables with the key set to table name.
			
		"""
		self.logger.debug("getting table metadata")
		table_include=""
		if self.mysql_con.tables_limit:
			self.logger.info("table copy limited to tables: %s" % ','.join(self.mysql_con.tables_limit))
			table_include="AND table_name IN ('"+"','".join(self.mysql_con.tables_limit)+"')"
		sql_tables="""
			SELECT 
				table_schema,
				table_name
			FROM 
				information_schema.TABLES 
			WHERE 
					table_type='BASE TABLE' 
				AND table_schema=%s
				"""+table_include+"""
			;
		"""
		
		self.mysql_con.my_cursor.execute(sql_tables, (self.mysql_con.my_database))
		table_list=self.mysql_con.my_cursor.fetchall()
		for table in table_list:
			column_data=self.get_column_metadata(table["table_name"])
			index_data=self.get_index_metadata(table["table_name"])
			dic_table={'name':table["table_name"], 'columns':column_data,  'indices': index_data}
			self.my_tables[table["table_name"]]=dic_table
		
	def print_progress (self, iteration, total, table_name):
		"""
			Print the copy progress. 
			As the variable total on innodb is estimated the percentage progress exceed the 100%.
			In order to reduce noise when the log level is info only the tables copied in multiple slices
			get the print progress.
			
			:param iteration: The slice number currently processed
			:param total: The estimated total slices
			:param table_name: The table name
		"""
		if total>1:
			self.logger.info("Table %s copied %d %%" % (table_name, 100 * float(iteration)/float(total)))
		else:
			self.logger.debug("Table %s copied %d %%" % (table_name, 100 * float(iteration)/float(total)))
		
	def generate_select(self, table_columns, mode="csv"):
		"""
			The method builds the select list using the dictionary table_columns which is the columns key in the table's metadata.
			The method can build a select list for a CSV output or an INSERT output. The default mode is csv.
			
			:param table_columns: The table's column dictionary with the column metadata
			:param mode: the select mode, csv or insert
		"""
		column_list=[]
		columns=""
		if mode=="csv":
			for column in table_columns:
					column_list.append("COALESCE(REPLACE("+column["column_csv"]+", '\"', '\"\"'),'NULL') ")
			columns="REPLACE(CONCAT('\"',CONCAT_WS('\",\"',"+','.join(column_list)+"),'\"'),'\"NULL\"','NULL')"
		if mode=="insert":
			for column in table_columns:
				column_list.append(column["column_select"])
			columns=','.join(column_list)
		return columns
	

	def insert_table_data(self, pg_engine, ins_arg):
		"""
			This method is a fallback procedure whether copy_table_data fails.
			The ins_args is a list with the informations required to run the select for building the insert
			statements and the slices's start and stop.
			The process is performed in memory and can take a very long time to complete.
			
			:param pg_engine: the postgresql engine
			:param ins_arg: the list with the insert arguments (slice_insert, table_name, columns_insert, slice_size)
		"""
		slice_insert=ins_arg[0]
		table_name=ins_arg[1]
		columns_ins=ins_arg[2]
		copy_limit=ins_arg[3]
		for slice in slice_insert:
			sql_out="SELECT "+columns_ins+"  FROM "+table_name+" LIMIT "+str(slice*copy_limit)+", "+str(copy_limit)+";"
			self.mysql_con.my_cursor_fallback.execute(sql_out)
			insert_data =  self.mysql_con.my_cursor_fallback.fetchall()
			pg_engine.insert_data(table_name, insert_data , self.my_tables)

	def copy_table_data(self, pg_engine,  copy_max_memory, lock_tables=True):
		"""
			copy the table data from mysql to postgres
			param pg_engine: The postgresql engine required to write into the postgres database.
			The process determines the estimated optimal slice size using copy_max_memory and avg_row_length 
			from MySQL's information_schema.TABLES. If the table contains no rows then the slice size is set to a
			reasonable high value (100,000) in order to get the table copied in one slice. The estimated numer of slices is determined using the 
			slice size.
			Then generate_select is used to build the csv and insert columns for the table.
			An unbuffered cursor is used to pull the data from MySQL using the CSV format. The fetchmany with copy_limit (slice size) is called
			to pull out the rows into a file object. 
			The copy_mode determines wheter to use a file (out_file) or an in memory file object (io.StringIO()).
			If there no more rows the loop exits, otherwise continue to the next slice. When the slice is saved the method pg_engine.copy_data is
			executed to load the data into the PostgreSQL table.
			If some error occurs the slice number is saved into the list slice_insert and after all the slices are copied the fallback procedure insert_table_data
			process the remaining slices using the inserts.
			
			:param pg_engine: the postgresql engine
			:param copy_max_memory: The estimated maximum amount of memory to use in a single slice copy
			:param lock_tables: Specifies whether the tables should be locked before copying the data
			
		"""
		out_file='%s/output_copy.csv' % self.out_dir
		self.logger.info("locking the tables")
		if lock_tables:
			self.lock_tables()
		table_list = []
		if pg_engine.table_limit[0] == '*':
			for table_name in self.my_tables:
				table_list.append(table_name)
		else:
			table_list = pg_engine.table_limit
		for table_name in table_list:
			slice_insert=[]
			
			try:
				table=self.my_tables[table_name]
				self.logger.info("copying table %s" %(table_name))
				table_name=table["name"]
				table_columns=table["columns"]
				self.logger.debug("estimating rows in "+table_name)
				sql_count=""" 
					SELECT 
						table_rows,
						CASE
							WHEN avg_row_length>0
							then
								round(("""+str(copy_max_memory)+"""/avg_row_length))
						ELSE
							0
						END as copy_limit
					FROM 
						information_schema.TABLES 
					WHERE 
							table_schema=%s 
						AND	table_type='BASE TABLE'
						AND table_name=%s 
					;
				"""
				self.mysql_con.my_cursor.execute(sql_count, (self.mysql_con.my_database, table_name))
				count_rows=self.mysql_con.my_cursor.fetchone()
				total_rows=count_rows["table_rows"]
				copy_limit=int(count_rows["copy_limit"])
				if copy_limit == 0:
					copy_limit=1000000
				num_slices=int(total_rows//copy_limit)
				range_slices=list(range(num_slices+1))
				total_slices=len(range_slices)
				slice=range_slices[0]
				self.logger.debug("%s will be copied in %s slices of %s rows"  % (table_name, total_slices, copy_limit))
				columns_csv=self.generate_select(table_columns, mode="csv")
				columns_ins=self.generate_select(table_columns, mode="insert")
				csv_data=""
				sql_out="SELECT "+columns_csv+" as data FROM `"+table_name+"`;"
				self.mysql_con.connect_db_ubf()
				try:
					self.logger.debug("Executing query for table %s"  % (table_name, ))
					self.mysql_con.my_cursor_ubf.execute(sql_out)
				except:
					self.logger.error("error when pulling data from %s. sql executed: %s" % (table_name, sql_out))
				
				self.logger.debug("Starting extraction loop for table %s"  % (table_name, ))
				while True:
					csv_results = self.mysql_con.my_cursor_ubf.fetchmany(copy_limit)
					if len(csv_results) == 0:
						break
					csv_data="\n".join(d[0] for d in csv_results )
					
					if self.mysql_con.copy_mode=='direct':
						csv_file=io.StringIO()
						csv_file.write(csv_data)
						csv_file.seek(0)

					if self.mysql_con.copy_mode=='file':
						csv_file=codecs.open(out_file, 'wb', self.mysql_con.my_charset)
						csv_file.write(csv_data)
						csv_file.close()
						csv_file=open(out_file, 'rb')
						
					try:
						pg_engine.copy_data(table_name, csv_file, self.my_tables)
					except:
						self.logger.info("table %s error in PostgreSQL copy, saving slice number for the fallback to insert statements " % (table_name, ))
						slice_insert.append(slice)
						
					self.print_progress(slice+1,total_slices, table_name)
					slice+=1
					csv_file.close()
				self.mysql_con.disconnect_db_ubf()
				if len(slice_insert)>0:
					ins_arg=[]
					ins_arg.append(slice_insert)
					ins_arg.append(table_name)
					ins_arg.append(columns_ins)
					ins_arg.append(copy_limit)
					self.insert_table_data(pg_engine, ins_arg)
			except:
				self.logger.info("There was an error when processing the table %s" %(table_name))

		if lock_tables:
			self.unlock_tables()
		try:
			remove(out_file)
		except:
			pass
			
	def get_master_status(self):
		"""
			The method gets the master's coordinates using the command SHOW MASTER STATUS.
			The dictionary cursor is stored in the class variable self.master_status
		"""
		t_sql_master="SHOW MASTER STATUS;"
		self.mysql_con.my_cursor.execute(t_sql_master)
		self.master_status=self.mysql_con.my_cursor.fetchall()		
		
		
	def lock_tables(self):
		""" 
			The method locks the tables using FLUSH TABLES WITH READ LOCK. The 
			tables locked are limited to the tables found by get_table_metadata.
			After locking the tables the metod gets the master's coordinates with get_master_status.
		"""
		self.locked_tables=[]
		for table_name in self.my_tables:
			table=self.my_tables[table_name]
			self.locked_tables.append(table["name"])
		t_sql_lock="FLUSH TABLES `"+"`, `".join(self.locked_tables)+"` WITH READ LOCK;"
		self.mysql_con.my_cursor.execute(t_sql_lock)
		self.get_master_status()
	
	def unlock_tables(self):
		"""
			The method unlocks the tables previously locked by lock_tables
		"""
		t_sql_unlock="UNLOCK TABLES;"
		self.mysql_con.my_cursor.execute(t_sql_unlock)
	
	
		
			
	def __del__(self):
		"""
			Class destructor, disconnects the mysql connections.
		"""
		self.mysql_con.disconnect_db()



from pg_chameleon import mysql_engine, pg_engine
import yaml
import sys
import os
import time
import logging
from logging.handlers  import TimedRotatingFileHandler
from tabulate import tabulate
from distutils.sysconfig import get_python_lib
from shutil import copy
import threading

class config_dir(object):
	""" 
		Class used to setup the local user configuration directory.
		The class constructor sets only the class variables for the method set_config.
		The function get_python_lib() is used to determine the python library where pg_chameleon is installed.
	"""
	def __init__(self):
		"""
			Class constructor.
		"""
		python_lib=get_python_lib()
		cham_dir = "%s/.pg_chameleon" % os.path.expanduser('~')	
		local_config = "%s/config/" % cham_dir 
		local_logs = "%s/logs/" % cham_dir 
		local_pid = "%s/pid/" % cham_dir 
		self.global_config_example = '%s/pg_chameleon/config/config-example.yaml' % python_lib
		self.local_config_example = '%s/config-example.yaml' % local_config
		self.conf_dirs=[
			cham_dir, 
			local_config, 
			local_logs, 
			local_pid, 
			
		]
		
	def set_config(self):
		""" 
			The method loops the list self.conf_dirs creating it only if missing.
			
			The method checks the freshness of the config-example.yaml file and copies the new version
			from the python library determined in the class constructor with get_python_lib().
			
			If the configuration file is missing the method copies the file with a different message.
		
		"""
		for confdir in self.conf_dirs:
			if not os.path.isdir(confdir):
				print ("creating directory %s" % confdir)
				os.mkdir(confdir)
		
		if os.path.isfile(self.local_config_example):
			if os.path.getctime(self.global_config_example)>os.path.getctime(self.local_config_example):
				print ("updating config_example %s" % self.local_config_example)
				copy(self.global_config_example, self.local_config_example)
		else:
			print ("copying config_example %s" % self.local_config_example)
			copy(self.global_config_example, self.local_config_example)
		
			
			
			

class global_config(object):
	"""
		This class parses the configuration file specified by the parameter config_name and sets 
		the class variables used by the replica_engine class. 
		The constructor checks if the configuration file is present and if is missing emits an error message followed by
		the sys.exit() command. If the configuration file is successful parsed the class variables are set from the
		configuration values. 
		
		The  function get_python_lib() is used to determine the library directory where pg_chameleon is installed in order to get the
		sql files.
		The configuration files are searched in the $HOME/.pg_chameleon/config.
		Should any parameter be missing in config the class constructor emits an error message specifying the parameter with reference to config-example.yaml.
		
		:param config_name: the configuration file to use. If omitted is set to default.
	
	"""
	def __init__(self,config_name="default", debug_mode=False):
		"""
			The class  constructor.
		"""
		python_lib=get_python_lib()
		cham_dir = "%s/.pg_chameleon" % os.path.expanduser('~')	
		config_dir = '%s/config/' % cham_dir
		sql_dir = "%s/pg_chameleon/sql/" % python_lib
		
		
		if os.path.isdir(sql_dir):
				self.sql_dir = sql_dir
		else:
			print("**FATAL - sql directory %s missing "  % self.sql_dir)
			sys.exit(1)
			
		config_file = '%s/%s.yaml' % (config_dir, config_name)
		if os.path.isfile(config_file):
			self.config_name = config_name
			self.config_dir = config_dir
		else:
			print("**FATAL - could not find the configuration file %s.yaml in %s"  % (config_name, config_dir))
			sys.exit(2)
		conffile = open(config_file, 'rb')
		confdic = yaml.load(conffile.read())
		conffile.close()
		try:
			self.source_name = confdic["source_name"]
			self.dest_schema = confdic["dest_schema"]
			self.mysql_conn = {key: str(confdic["mysql_conn"][key]) for key in confdic["mysql_conn"] if key != "port"}
			self.pg_conn = {key: str(confdic["pg_conn"][key]) for key in confdic["pg_conn"]}
			self.my_database = confdic["my_database"]
			self.my_charset = confdic["my_charset"]
			self.pg_charset = confdic["pg_charset"]
			self.pg_database = confdic["pg_database"]
			self.my_server_id = confdic["my_server_id"]
			self.replica_batch_size = confdic["replica_batch_size"]
			self.tables_limit = confdic["tables_limit"]
			self.copy_mode = confdic["copy_mode"]
			self.hexify = confdic["hexify"]
			self.log_level = confdic["log_level"]
			self.log_dest = confdic["log_dest"]
			self.log_days_keep = confdic["log_days_keep"]
			self.out_dir = confdic["out_dir"]
			self.type_override = confdic["type_override"]
			
			self.sleep_loop = confdic["sleep_loop"]
			self.pause_on_reindex = confdic["pause_on_reindex"]
			self.sleep_on_reindex = confdic["sleep_on_reindex"]
			self.reindex_app_names = confdic["reindex_app_names"]
			self.batch_retention = confdic["batch_retention"]
			
			self.log_file = os.path.expanduser(confdic["log_dir"])+config_name+'.log'
			self.pid_file = os.path.expanduser(confdic["pid_dir"])+"/"+config_name+".pid"
			self.exit_file = os.path.expanduser(confdic["pid_dir"])+"/"+config_name+".lock"
			copy_max_memory = str(confdic["copy_max_memory"])[:-1]
			copy_scale = str(confdic["copy_max_memory"])[-1]
			try:
				int(copy_scale)
				copy_max_memory = confdic["copy_max_memory"]
			except:
				if copy_scale =='k':
					copy_max_memory = str(int(copy_max_memory)*1024)
				elif copy_scale =='M':
					copy_max_memory = str(int(copy_max_memory)*1024*1024)
				elif copy_scale =='G':
					copy_max_memory = str(int(copy_max_memory)*1024*1024*1024)
				else:
					print("**FATAL - invalid suffix in parameter copy_max_memory  (accepted values are (k)ilobytes, (M)egabytes, (G)igabytes.")
					sys.exit(3)
			self.copy_max_memory = copy_max_memory
		except KeyError as key_missing:
			print('Missing key %s in configuration file. check %s/config-example.yaml for reference' % (key_missing, self.config_dir))
			sys.exit(4)
	
	
	def get_source_name(self, config_name = 'default'):
		"""
		The method tries to set the parameter source_name determined from the configuration file.
		The value is used to query the replica catalog in order to get the source sstatus in method list_config().
		
		:param config_name: the configuration file to use. If omitted is set to default.
		"""
		
		config_file = '%s/%s.yaml' % (self.config_dir, config_name)
		self.config_name = config_name
		if os.path.isfile(config_file):
			conffile = open(config_file, 'rb')
			confdic = yaml.load(conffile.read())
			conffile.close()
			try:
				source_name=confdic["source_name"]
			except:
				print('FATAL - missing parameter source name in config file %s' % config_file)
				source_name='NOT CONFIGURED'
		return source_name
		
class replica_engine(object):
	"""
		This class is wraps the the mysql and postgresql engines in order to perform the various activities required for the replica. 
		The constructor inits the global configuration class  and setup the mysql and postgresql engines as class objects. 
		The class sets the logging using the configuration parameter.
		
	"""
	def __init__(self, config, debug_mode=False, nolock=False):
		"""
			Class constructor
			:param stdout: forces the logging to stdout even if the logging destination is file
		"""
		self.debug_mode = debug_mode
		self.nolock=nolock
		self.lst_yes= ['yes',  'Yes', 'y', 'Y']
		self.global_config=global_config(config)
		self.logger = logging.getLogger(__name__)
		self.logger.setLevel(logging.DEBUG)
		self.logger.propagate = False
		formatter = logging.Formatter("%(asctime)s %(threadName)s %(levelname)s %(filename)s (%(lineno)s): %(message)s", "%b %e %H:%M:%S")
		
		if self.global_config.log_dest=='stdout' or self.debug_mode:
			fh=logging.StreamHandler(sys.stdout)
			
		elif self.global_config.log_dest=='file':
			fh = TimedRotatingFileHandler(self.global_config.log_file, when="d",interval=1,backupCount=self.global_config.log_days_keep)
		
		if self.global_config.log_level=='debug' or self.debug_mode:
			fh.setLevel(logging.DEBUG)
		elif self.global_config.log_level=='info':
			fh.setLevel(logging.INFO)
			
		fh.setFormatter(formatter)
		self.logger.addHandler(fh)

		self.my_eng=mysql_engine(self.global_config, self.logger)
		self.pg_eng=pg_engine(self.global_config, self.my_eng.my_tables, self.my_eng.table_file, self.logger, self.global_config.sql_dir)
		self.sleep_loop=self.global_config.sleep_loop
		
		self.pid_file = self.global_config.pid_file
		self.exit_file = self.global_config.exit_file
		self.source_schema = self.global_config.my_database
		self.pg_eng.source_schema = self.source_schema
	
	def detach_replica(self):
		"""
			The method terminates the replica, remove the source from the table t_sources and resets the sequences in 
			the postgresql database, leaving the replica snapshot capable to continue the activity directly on PostgreSQL
		"""
		source_name = self.global_config.source_name
		drp_msg = 'Detaching the replica will remove any reference for the source %s.\n Are you sure? YES/No\n'  % source_name
		fk_metadata = self.my_eng.get_fk_metadata()
			
		if sys.version_info[0] == 3:
			drop_src = input(drp_msg)
		else:
			drop_src = raw_input(drp_msg)
		if drop_src == 'YES':
			self.stop_replica(allow_restart=False)
			self.pg_eng.reset_sequences(self.global_config.source_name)
			self.pg_eng.add_foreign_keys(self.global_config.source_name, fk_metadata)
			self.pg_eng.drop_source(self.global_config.source_name)
		elif drop_src in  self.lst_yes:
			print('Please type YES all uppercase to confirm')
		sys.exit()
		
		print('replica detached')
	
	
	def init_replica(self):
		"""
			The method initialise a replica.
		
			It calls the pg_engine methods set_source_id which sets the source identifier and change the source status.
			
			The pg_engine method clean_batch_data is used to remove any unreplayed row in the tables t_log_replica_1(2).
			
			The class methods create_schema, copy_table_data and create_indices are called in sequence to initialise the replica.
			
		"""
		self.pg_eng.set_source_id('initialising')
		self.pg_eng.clean_batch_data()
		self.create_schema()
		self.copy_table_data()
		self.create_indices()
		self.pg_eng.set_source_id('initialised')

	def wait_for_replica_end(self):
		""" 
			The method is used to wait for the replica's end monitoring the pid file.
			The replica status is determined using the check_running method passing the write_pid=false.
			
			There is a 5 seconds sleep between each check.
			
		"""
		self.logger.info("waiting for replica process to stop")
		while True:
			replica_running=self.check_running(write_pid=False)
			if not replica_running:
				break
			time.sleep(5)
		
		self.logger.info("replica process stopped")
	

	def stop_replica(self, allow_restart=True):
		"""
			the method writes the exit file in the pid directory and waits for the replica process's end.
			If allow_restart is true the exit file is removed.
			
			:param allow_restart: determines whether the exit file is removed or not in order to allow the replica to start again.
		"""
		exit=open(self.exit_file, 'w')
		exit.close()
		self.wait_for_replica_end()
		if allow_restart:
			os.remove(self.exit_file)
	
	def enable_replica(self):
		"""
			The  method removes the exit file in order to let the replica start again.
		"""
		try:
			os.remove(self.exit_file)
			self.logger.info("Replica enabled")
		except:
			self.logger.info("Replica already enabled")
			
	def  create_schema(self):
		"""
			Creates the database schema on PostgreSQL using the metadata extracted from MySQL.
		"""
		self.pg_eng.create_schema()
		self.logger.info("Importing mysql schema")
		self.pg_eng.build_tab_ddl()
		self.pg_eng.create_tables()
	

	
	def  create_indices(self):
		"""
			Creates the indices on the PostgreSQL schema using the metadata extracted from MySQL.
			
		"""
		self.pg_eng.build_idx_ddl()
		self.pg_eng.create_indices()
	
	def create_service_schema(self):
		"""
			Creates the service schema sch_chameleon on the PostgreSQL database. The service schema is required for having the replica working correctly.
	
		"""
		self.pg_eng.create_service_schema()
		
	def upgrade_service_schema(self):
		"""
			Upgrade the service schema to the latest version. 
			
		"""
		self.pg_eng.upgrade_service_schema()
		
	def drop_service_schema(self):
		"""
			Drops the service schema. The action discards any replica data.

		"""
		self.logger.info("Dropping the service schema")
		self.pg_eng.drop_service_schema()
		
	def check_file_exit(self):
		process_exit=False
		"""checks for the exit file and terminate the replica if the file is present. If there is the pid file is removed before  the function's return """
		if os.path.isfile(self.exit_file):
			if os.path.isfile(self.pid_file):
				self.logger.info("exit file detected, removing the pid file and terminating the replica process")
				os.remove(self.pid_file)
			else:
				self.logger.info("you shall remove the file %s before starting the replica process " % self.exit_file)
			process_exit=True
		return process_exit
	
	
	def check_running(self, write_pid=False):
		""" 
			checks if the process is running. 
			
			:param write_pid: determines whether the pid file is written or not. Used if we just need to check if the replica is running.
		"""
		
		process_running=False 
		try:
			file_pid=open(self.pid_file,'r')
			pid=file_pid.read()
			file_pid.close()
			os.kill(int(pid),0)
			process_running=True
		except:
			if write_pid:
				pid=os.getpid()
				file_pid=open(self.pid_file,'w')
				file_pid.write(str(pid))
				file_pid.close()
			process_running=False
		return process_running
	
	def read_replica(self):
		while True:
			try:
				self.my_eng.run_replica(self.pg_eng)
			except:
				break
			
			exit_request = self.check_file_exit()
			if exit_request:
				break
			time.sleep(self.sleep_loop)
			
	def replay_replica(self):
		while True:
			try:
				self.pg_eng.process_batch(self.global_config.replica_batch_size)
			except:
				break
			exit_request = self.check_file_exit()
			if exit_request:
				break
			time.sleep(self.sleep_loop)
			
	def run_replica_thread(self):
		"""
			Threaded version of run replica.
		"""
		replica_possible = self.my_eng.check_mysql_config()
		if replica_possible:
			self.logger.info("Configuration on MySQL allows replica.")
		else:
			print("** FATAL - The mysql configuration do not allow the replica.\n The parameters log_bin, binlog_format  and binlog_row_image are not set correctly.\n Check the documentation for further details.\n http://www.pgchameleon.org/documents/")
			sys.exit()
		already_running = self.check_running(write_pid=True)
		exit_request = self.check_file_exit()
		
		if already_running:
			sys.exit()
		if exit_request:
			self.pg_eng.set_source_id('stopped')
			sys.exit()
		self.pg_eng.set_source_id('running')
		read_replica = threading.Thread(target=self.read_replica, name='read_replica')
		read_replica.setDaemon(True)
		replay_replica = threading.Thread(target=self.replay_replica, name='replay_replica')
		replay_replica.setDaemon(True)
		read_replica.start()
		replay_replica.start()
		while True:
			read_alive = read_replica.isAlive()
			replay_alive = replay_replica.isAlive()
			self.logger.info("Read thread running: %s - Replay thread running: %s" %(read_alive, replay_alive))
			if not read_alive and not replay_alive:
				self.pg_eng.set_source_id('stopped')
				break
			if (read_alive and not replay_alive) or (not read_alive and replay_alive):
				self.stop_replica()
				break
			time.sleep(self.sleep_loop)
	
	
	def run_replica(self):
		"""
			The method starts and run the replica loop. 
			Before starting the loop checks if the replica is already running with check_running with write_pid=True.
			
			It also checks whether the exit file is present or not. 
			If present skips the replica start, otherwise start the while loop which runs the mysql_engine method run_replica.
			
			When the mysql_engine.run_replica() completes it checks if there is the exit file and eventually exit the loop.
			Otherwise sleeps for the amount or seconds set in sleep_loop.
			
		"""
		replica_possible = self.my_eng.check_mysql_config()
		if replica_possible:
			self.logger.info("Configuration on MySQL allows replica.")
		else:
			print("** FATAL - The mysql configuration do not allow the replica.\n The parameters log_bin, binlog_format  and binlog_row_image are not set correctly.\n Check the documentation for further details.\n http://www.pgchameleon.org/documents/")
			sys.exit()
		already_running = self.check_running(write_pid=True)
		exit_request = self.check_file_exit()
		
		if already_running:
			self.logger.error("The replica is already running")
			sys.exit()
		if exit_request:
			self.pg_eng.set_source_id('stopped')
			sys.exit()
		
		self.pg_eng.set_source_id('running')
		
		while True:
			if self.debug_mode or self.nolock:
				self.my_eng.run_replica(self.pg_eng)
				self.pg_eng.process_batch(self.global_config.replica_batch_size)
			else:
				try:
					self.my_eng.run_replica(self.pg_eng)
					self.pg_eng.process_batch(self.global_config.replica_batch_size)
				except :
					self.pg_eng.set_source_id('error')
					self.logger.error("An error occurred during the replica. %s" % (sys.exc_info(), ))
					exit=open(self.exit_file, 'w')
					exit.close()
					raise
					sys.exit(5)
			self.logger.info("batch complete. sleeping %s second(s)" % (self.sleep_loop, ))
			if self.check_file_exit():
				break
			time.sleep(self.sleep_loop)
		self.pg_eng.set_source_id('stopped')
	
	def list_config(self):
		"""
			List the available configurations stored in ~/.pg_chameleon/config/
		"""
		list_config = (os.listdir(self.global_config.config_dir))
		tab_headers = ['Config file',  'Source name',  'Status']
		tab_body = []
		
		for file in list_config:
			lst_file = file.split('.')
			file_name = lst_file[0]
			file_ext = lst_file[1]
			if file_ext == 'yaml' and file_name!='config-example':
				source_name = self.global_config.get_source_name(file_name)
				source_status = self.pg_eng.get_source_status(source_name)
				tab_row = [file_name, source_name, source_status]
				tab_body.append(tab_row)
		print(tabulate(tab_body, headers=tab_headers))
	
	def show_status(self):
		"""
			list the replica status using the configuration files and the replica catalogue
		"""
		source_status=self.pg_eng.get_status()
		tab_headers = ['Config file',  'Source schema', 'Destination schema',  'Status' ,  'Read lag',  'Last read',  'Replay lag' , 'Last replay']
		tab_body = []
			
		for status in source_status:
			source_name = status[0]
			dest_schema = status[1]
			source_status = status[2]
			read_lag = status[3]
			last_read = status[4]
			replay_lag = status[5]
			last_replay = status[6]
			source_schema = status[7]
			tab_row = [source_name, source_schema,  dest_schema, source_status, read_lag, last_read,  replay_lag, last_replay]
			tab_body.append(tab_row)
		print(tabulate(tab_body, headers=tab_headers))
		
	def add_source(self):
		"""
			register the configuration source in the replica catalogue
		"""
		source_name = self.global_config.source_name
		dest_schema = self.global_config.dest_schema
		self.pg_eng.add_source(source_name, dest_schema)

	def drop_source(self):
		"""
			remove the configuration source and all the replica informations associated with the source from the replica catalogue
		"""
		source_name = self.global_config.source_name
		drp_msg = 'Dropping the source %s will remove drop any replica reference.\n Are you sure? YES/No\n'  % source_name
		if sys.version_info[0] == 3:
			drop_src = input(drp_msg)
		else:
			drop_src = raw_input(drp_msg)
		if drop_src == 'YES':
			self.pg_eng.drop_source(self.global_config.source_name)
		elif drop_src in  self.lst_yes:
			print('Please type YES all uppercase to confirm')
		sys.exit()
		
	def copy_table_data(self, truncate_tables=False):
		"""
			Copy the data for the replicated tables from mysql to postgres.
			After the copy the master's coordinates are saved in postgres.
			
			:param truncate_tables: determines whether the existing tables should be truncated before running a copy table data
		"""
		if truncate_tables:
			self.pg_eng.truncate_tables()
		self.my_eng.copy_table_data(self.pg_eng, self.global_config.copy_max_memory)
		self.pg_eng.save_master_status(self.my_eng.master_status, cleanup=True)

	def sync_tables(self, table):
		"""
			syncronise single tables with the mysql master.
			The process attempts to drop the existing tables or add them if not present.
			The tables are stored in the replica catalogue with their master's coordinates and are ignored until the replica process reaches
			the correct position. 
			:param table: comma separated list of table names to synchronise
		"""
		if table != "*":
			table_limit = table.split(',')
			self.my_eng.mysql_con.tables_limit = table_limit
			self.my_eng.get_table_metadata()
			self.my_eng.lock_tables()
			self.pg_eng.table_limit = table_limit
			self.pg_eng.master_status = self.my_eng.master_status
			self.pg_eng.set_source_id('initialising')
			self.stop_replica(allow_restart=False)
			self.pg_eng.build_tab_ddl()
			self.pg_eng.drop_tables()
			self.pg_eng.create_tables()
			self.my_eng.copy_table_data(self.pg_eng,  self.global_config.copy_max_memory, False)
			self.pg_eng.create_indices()
			self.pg_eng.delete_table_events()
			self.pg_eng.set_source_id('initialised')
			self.my_eng.unlock_tables()
			self.enable_replica()
		else:
			print("You should specify at least one table to synchronise.")

